
"""
   Job Base class
"""

from abc import ABC

class JobBase(ABC):
    """
        Create a Job object

        :param agent: agent object that created the listing.
        :type agent: :class:`.Agent` object to assign to this Listing
        :param str job_id: id of the job.
    """

    IsWorkingStatusList = [
        'scheduled',
        'running',
        'accepted'
    ]
    def __init__(self, job_id, status=None, results=None):
        """

        init the the Job Object Base with the agent instance

        :param int job_id: id of the job
        :param str status: status of the job
        :param dict results: dict or None for the results
        """
        self._job_id = job_id
        self._status = status
        self._results = results
        super().__init__()

    @property
    def is_finished(self):
        if self._status:
            return self._status not in JobBase.IsWorkingStatusList
        return False

    @property
    def job_id(self):
        """
        Return the job id

        :return: Job id
        :type: int
        """
        return self._job_id

    @property
    def status(self):
        """
        Return the status for this job

        :return: status
        :type: str
        """
        return self._status

    @property
    def results(self):
        """
        Return the results for this job

        :return: results
        :type: dict or None
        """
        return self._results

    def __str__(self):
        text = 'job:'
        if self._status:
            text += f'{self._status}'
        if self._results:
            text += f' = {self._results}'
        return text



"""
    test_20_invoke_service_free

    As a developer working with Ocean,
    I wish to invoke a free service available on the Ocean ecosystem and obtain the results as a new Ocean Asset

"""

import secrets
import logging
import json
import time

from starfish.asset import OperationAsset
from starfish.job import Job

PRIME_NUMBER_OPERATION_ASSET_ID = "0x0e48ad0c07f6fe87762e24cba3e013a029b7cd734310bface8b3218280366791"
TO_HASH_OPERATION_ASSET_ID = "0x678d5e333ca9ea1a0f7939b4f1d923f73a1641dda8da0430c2b3604d3ceb5991"

TEST_HASH_TEXT = 'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Vestibulum eu congue odio, vel congue sapien. Morbi ac purus ornare, volutpat elit a, lacinia odio. Integer tempor tellus eget iaculis lacinia. Curabitur aliquam, dui vel vestibulum rhoncus, enim metus interdum enim, in sagittis massa est vel velit. Nunc venenatis commodo libero, vitae elementum nulla ultricies id. Aliquam erat volutpat. Cras eu pretium lacus, quis facilisis mauris. Duis sem quam, blandit id tempor in, porttitor at neque. Cras ut blandit risus. Maecenas vitae sodales neque, eu ultricies nibh.'

def test_20_prime_number_sync(remote_agent):

    # this should be get_listing..
    # listing = remote_agent.get_listing(listing_id)

    # at the moment Koi does not create a listing with the asset, so
    # we need to call get_asset instead
    operation_asset = remote_agent.get_asset(PRIME_NUMBER_OPERATION_ASSET_ID)
    assert(operation_asset)
    assert(isinstance(operation_asset, OperationAsset))
    assert(operation_asset.metadata['type'] == 'operation')

    params = {
        'first-n': '11'
    }

    response = remote_agent.invoke_result(operation_asset, params)
    assert(response)
    assert(response['results'])
    print(response)

def test_20_prime_number_async(remote_agent):

    # this should be get_listing..
    # listing = remote_agent.get_listing(listing_id)

    # at the moment Koi does not create a listing with the asset, so
    # we need to call get_asset instead
    operation_asset = remote_agent.get_asset(PRIME_NUMBER_OPERATION_ASSET_ID)
    assert(operation_asset)
    assert(isinstance(operation_asset, OperationAsset))
    assert(operation_asset.metadata['type'] == 'operation')

    params = {
        'first-n': '11'
    }

    response = remote_agent.invoke_result(operation_asset, params, True)
    assert(response)
    assert(response['jobid'])

    job_id = int(response['jobid'])
    assert(isinstance(job_id, int))

    # FIXME: bug in koi, can return an empty job status record, straight after job creation
    time.sleep(1)
    # test get_job
    job = remote_agent.get_job(job_id)
    assert(job)
    assert(isinstance(job, Job))

    # test wait_for_job
    job = remote_agent.job_wait_for_completion(job_id)
    assert(job)
    assert(isinstance(job, Job))
    assert(job.status == 'succeeded' or job.status == 'completed')
    assert(job.results)
    print(job)



def test_20_to_hash_sync(remote_agent):

    # this should be get_listing..
    # listing = remote_agent.get_listing(listing_id)

    # at the moment Koi does not create a listing with the asset, so
    # we need to call get_asset instead
    operation_asset = remote_agent.get_asset(TO_HASH_OPERATION_ASSET_ID)
    assert(operation_asset)
    assert(isinstance(operation_asset, OperationAsset))
    assert(operation_asset.metadata['type'] == 'operation')

    params = {
        'to-hash': TEST_HASH_TEXT
    }

    response = remote_agent.invoke_result(operation_asset, params)
    assert(response)
    assert(response['results'])
    print(response)

def test_20_to_hash_async(remote_agent):

    # this should be get_listing..
    # listing = remote_agent.get_listing(listing_id)

    # at the moment Koi does not create a listing with the asset, so
    # we need to call get_asset instead
    operation_asset = remote_agent.get_asset(TO_HASH_OPERATION_ASSET_ID)
    assert(operation_asset)
    assert(isinstance(operation_asset, OperationAsset))
    assert(operation_asset.metadata['type'] == 'operation')

    params = {
        'to-hash': TEST_HASH_TEXT
    }

    response = remote_agent.invoke_result(operation_asset, params, True)
    assert(response)
    assert(response['jobid'])

    job_id = int(response['jobid'])
    assert(isinstance(job_id, int))

    # FIXME: bug in koi, can return an empty job status record, straight after job creation
    time.sleep(1)

    # test get_job
    job = remote_agent.get_job(job_id)
    assert(job)
    assert(isinstance(job, Job))


    # test wait_for_job
    job = remote_agent.job_wait_for_completion(job_id)
    assert(job)
    assert(isinstance(job, Job))
    assert(job.status == 'completed')
    assert(job.results)
    print(job)




import pytest
import secrets


from starfish.agent.invoke_agent import InvokeAgent

TEST_KOI_URL = 'http:testkoi.com:8080'


def test_init(ocean):
    agent = InvokeAgent()
    assert(agent)
    
    agent = InvokeAgent(koi_url=TEST_KOI_URL)
    assert(agent)

def test_get_operations():
    # this class is not ready yet for 
    # unit testing 'requests'
    """
    agent = InvokeAgent(koi_url=TEST_KOI_URL)
    operations = agent.get_operations()
    assert(operations)
    """



"""

    Command line Test output

"""


from .base_output import (
    BaseOutput,
)

class PyTestOutput(BaseOutput):

    def __init__(self):
        BaseOutput.__init__(self)

    def show(self, text=None, name=None):
        if text:
            if name is None:
                name = 'value'
            self.show_item(name, text)

    def show_item(self, name, value, format_text=None):
        self._items[name] = value

    def show_header(self):
        pass

    def show_items(self):
        self._rows.append(self._items)
        self._items = {}

    def flush(self):
        pass



"""
    test_18_confirm_purchase

    As a developer building a service provider Agent for Ocean,
    I need a way to confirm if an Asset has been sucessfully puchased so that
    I can determine whether to serve the asset to a given requestor

"""

import secrets
import logging
import json

from starfish.asset import DataAsset


def test_18_confirm_purchase(ocean, resources, config, remote_agent, purchaser_account):
    test_data = secrets.token_hex(1024)
    asset = DataAsset.create('TestAsset', test_data)
    listing = remote_agent.register_asset(asset, resources.listing_data)
    listing.set_published(True)
    logging.debug("confirm_purchase for listingid: " + listing.listing_id)
    response = remote_agent.update_listing(listing)
    logging.debug("update_listing response: " + str(response))
    assert(response)
    status = 'ordered'
    purchase = remote_agent.purchase_asset(listing, purchaser_account, None, status)
    assert(purchase['listingid'] == listing.listing_id)
    assert(purchase['status'] == status)



"""
    test_10_asset_upload

    As a developer working with an Ocean marketplace,
    I need a way to upload my asset with a service agreement

"""

import secrets

from starfish.asset import DataAsset


def test_10_asset_upload(resources, remote_agent):
    test_data = secrets.token_hex(1024)
    asset = DataAsset.create('TestAsset', test_data)
    listing = remote_agent.register_asset(asset, resources.listing_data)
    assert(listing)
    assert(listing.asset)

    assert(remote_agent.upload_asset(listing.asset))



#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

import logging
import os

import sys

from squid_py import Config


def get_variable_value(variable):
    if os.getenv(variable) is None:
        logging.error(f'you should provide a {variable}')
        sys.exit(1)
    else:
        return os.getenv(variable)


class ExampleConfig:
    _local_aqua_url = "http://172.15.0.15:5000"
    _local_brizo_url = "http://localhost:8030"
    _duero_aqua_url = "https://aquarius.duero.dev-ocean.com"
    _duero_brizo_url = "https://brizo.duero.dev-ocean.com"
    # _nile_aqua_url = "http://172.15.0.15:5000"

    # _nile_aqua_url = "https://nginx-aquarius.dev-ocean.com"
    # _nile_brizo_url = "https://nginx-brizo.dev-ocean.com"
    # _nile_aqua_url = "https://nginx-aquarius.dev-ocean.com"
    _nile_aqua_url = "https://aquarius.marketplace.dev-ocean.com"
    # _nile_aqua_url = "http://172.15.0.15:5000"
    _nile_brizo_url = "https://brizo.marketplace.dev-ocean.com"
    # _nile_brizo_url = "http://localhost:8030"

    _duero_secret_store_url = "https://secret-store.duero.dev-ocean.com"
    _nile_secret_store_url = "https://secret-store.dev-ocean.com"
    # _nile_secret_store_url = "https://secret-store.marketplace.dev-ocean.com"
    _kovan_keeper_url = "http://localhost:8545"
    _remote_keeper_url = "https://%s.dev-ocean.com"
    _parity_url = "http://localhost:8545"
    _net_to_services_url = {
        'duero': {'aquarius': _duero_aqua_url, 'brizo': _duero_brizo_url},
        'nile': {'aquarius': _nile_aqua_url, 'brizo': _nile_brizo_url},
        'kovan': {'aquarius': _local_aqua_url, 'brizo': _local_brizo_url}
    }
    _net_name_map = {
        'duero': 'duero',
        'duero_local': 'duero',
        'nile': 'nile',
        'nile_local': 'nile',
        'kovan': 'kovan',
        'kovan_local': 'kovan'
    }
    _net_to_env_name = {
        'nile': 'TEST_NILE',
        'nile_local': 'TEST_LOCAL_NILE',
        'duero': 'TEST_DUERO',
        'duero_local': 'TEST_LOCAL_DUERO',
        'spree': 'TEST_LOCAL_SPREE',
        'kovan': 'TEST_KOVAN',
        'kovan_local': 'TEST_LOCAL_KOVAN'
    }

    @staticmethod
    def get_config_net():
        return os.environ.get('TEST_NET', 'spree')

    @staticmethod
    def get_env_name():
        net = ExampleConfig.get_config_net()
        return ExampleConfig._net_to_env_name.get(net)

    @staticmethod
    def get_accounts_config(local_node=True):
        if local_node:
            a, p, k = "0x00bd138abd70e2f00903268f3db08f2d25677c9e", "node0", "tests/resources/account_key_files/key_file_2.json"
            a1, p1, k1 = "0x068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0", "secret", "tests/resources/account_key_files/key_file_1.json"
        else:
            a, p, k = get_variable_value('PARITY_ADDRESS'), get_variable_value('PARITY_PASSWORD'), get_variable_value('PARITY_KEYFILE')
            a1, p1, k1 = get_variable_value('PARITY_ADDRESS1'), get_variable_value('PARITY_PASSWORD1'), get_variable_value('PARITY_KEYFILE1')

        return {
           "parity.address": a, "parity.password": p, "parity.keyfile": k,
           "parity.address1": a1, "parity.password1": p1, "parity.keyfile1": k1,
        }

    @staticmethod
    def get_base_config(local_node=True):
        config = {
            "keeper-contracts": {
                "keeper.url": "http://localhost:8545",
                "keeper.path": "artifacts",
                "secret_store.url": "http://localhost:12001",
                "parity.url": "http://localhost:8545",
            },
            "resources": {
                "aquarius.url": "http://localhost:5000",
                "brizo.url": "http://localhost:8030",
                "storage.path": "squid_py.db",
                "downloads.path": "consume-downloads"
            }
        }
        config['keeper-contracts'].update(ExampleConfig.get_accounts_config(local_node))
        return config

    @staticmethod
    def _get_config(local_node=True, net_key=''):
        config = ExampleConfig.get_base_config(local_node=local_node)
        net_name = ExampleConfig._net_name_map.get(net_key)
        if net_name == 'kovan':
            config['keeper-contracts']['keeper.url'] = ExampleConfig._kovan_keeper_url
        elif not local_node:
            config['keeper-contracts']['keeper.url'] = ExampleConfig._remote_keeper_url % net_name

        if net_name:
            config['keeper-contracts']['secret_store.url'] = \
                ExampleConfig._duero_secret_store_url if net_name == 'duero' \
                else ExampleConfig._nile_secret_store_url

            service_url = ExampleConfig._net_to_services_url[net_name]
            config['resources']['aquarius.url'] = service_url['aquarius']
            config['resources']['brizo.url'] = service_url['brizo']

        # parity_url maybe different than the keeper_url
        config['keeper-contracts']['parity.url'] = ExampleConfig._parity_url
        return config

    @staticmethod
    def get_config_dict():
        test_net = ExampleConfig.get_config_net()
        local_node = not test_net or test_net in ('nile_local', 'duero_local', 'spree', 'kovan_local')
        config_dict = ExampleConfig._get_config(local_node, test_net)
        return config_dict

    @staticmethod
    def get_config():
        logging.info("Configuration loaded for environment '{}'"
                     .format(ExampleConfig.get_config_net()))
        return Config(options_dict=ExampleConfig.get_config_dict())




import datetime
import pytest
import secrets
import tempfile


from starfish.agent.memory_agent import MemoryAgent
from starfish.asset.data_asset import DataAsset

VALID_DID = 'did:dep:' + secrets.token_hex(64)
INVALID_DID = 'did:dep:' + secrets.token_hex(128)


def _register_asset(ocean, resources, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    agent = MemoryAgent(ocean)
    assert(agent)
    asset = DataAsset.create('test memory agent asset', secrets.token_hex(1024))
    assert(asset)
    listing = agent.register_asset(asset, resources.listing_data, account)
    return (listing, agent, asset)

def _purchase_asset(ocean, resources, config):
    listing, agent, asset = _register_asset(ocean, resources, config)
    account = ocean.load_account(config.accounts[1].as_dict)
    purchase = agent.purchase_asset(listing, account)
    return purchase, listing, agent, asset, account

def test_init(ocean):
    agent = MemoryAgent(ocean)
    assert(agent)

def test_register_asset(ocean, resources, config):
    listing, agent, asset = _register_asset(ocean, resources, config)
    assert(listing)
    assert(listing.listing_id)

def test_get_listing(ocean, resources, config):
    listing, agent, asset = _register_asset(ocean, resources, config)
    found_listing = agent.get_listing(listing.listing_id)
    assert(found_listing)
    assert(found_listing.listing_id == listing.listing_id)

def test_search_listings(ocean, resources, config):
    listing, agent, asset = _register_asset(ocean, resources, config)
    listing_ids = agent.search_listings(resources.listing_data['author'])
    assert(listing_ids)
    assert(len(listing_ids) > 0)
    is_found = False
    for listing_id in listing_ids:
        if listing_id == listing.listing_id:
            is_found = True
            break
    assert(is_found)

def test_purchase_asset(ocean, resources, config):
    listing, agent, asset = _register_asset(ocean, resources, config)
    account = ocean.load_account(config.accounts[1].as_dict)
    purchase = agent.purchase_asset(listing, account)
    assert(purchase)


def test_is_access_granted_for_asset(ocean, resources, config):
    purchase, listing, agent, asset, account = _purchase_asset(ocean, resources, config)
    assert(agent.is_access_granted_for_asset(asset, account, purchase.purchase_id,))

def test_consume_asset(ocean, resources, config):
    purchase, listing, agent, asset, account = _purchase_asset(ocean, resources, config)
    assert(agent.consume_asset(listing, account, purchase.purchase_id,))

def test_is_did_valid():
    assert(MemoryAgent.is_did_valid(VALID_DID))
    # FIXME: This should be invalid
    assert( MemoryAgent.is_did_valid(INVALID_DID))



""""

    RemoteDataAsset:  DataAsset that has only a URL for assigned in the metadada

"""

from mimetypes import MimeTypes
from urllib.parse import urlparse

from starfish.asset.asset_base import AssetBase
from starfish.asset.data_asset import DataAsset

class RemoteDataAsset(DataAsset):

    @staticmethod
    def create_with_url(name, url, metadata=None, did=None):
        """

        Create a new RemoteDataAsset using a url.

        :param str name: Name of the asset to create
        :param str url: URL to assign to the asset
        :param dict metadata: Optional metadata to add to the assets metadata
        :param str did: Option DID to assign to this asset

        :return: a new DataAsset
        :type: :class:`.DataAsset`

        """

        metadata = AssetBase.generateMetadata(name, 'dataset', metadata)
        metadata['url'] = url
        if urlparse(url):
            metadata['contentType'] = 'application/octet-stream'
            mime = MimeTypes()
            mime_type = mime.guess_type(url)
            if mime_type and mime_type[0]:
                metadata['contentType'] = mime_type[0]

        return RemoteDataAsset(metadata, did)


    @property
    def url(self):
        if 'url' in self._metadata:
            return self._metadata['url']
        return None



"""

    Exceptions for starfish-py

"""


class OceanInvalidContractAddress(Exception):
    """  Raised when an invalid address is passed to the contract loader """

class OceanCommandLineError(Exception):
    """ raised on command line errors """

class StarfishPurchaseError(Exception):
    """ Raised when a purchase events have failed to complete """

class StarfishAssetNotFound(Exception):
    """ Raised when an asset is not found """

class StarfishAssetInvalid(Exception):
    """ Raised when a downloaded asset is not valid or has been changed """




"""
    Operation class
"""

import logging
from starfish.account import Account
from starfish.purchase.purchase_base import PurchaseBase
logger = logging.getLogger('starfish.squid_operation')

class SquidOperation(PurchaseBase):
    """

    This class is returned by purchasing an invokable asset uning the :func:`.Listing.purchase` method.

    :param agent: agent that was used create this object.
    :type agent: :class:`.Agent`
    :param listing: Listing used to purchase this asset.
    :type listing: :class:`.Listing`
    :param purchase_id: purchase_id used to buy this asset.
    :type purchase_id: str

    """

    def __init__(self, agent, listing, purchase_id):
        """init the the Operation Object Base with the agent instance"""
        super().__init__(agent, listing, purchase_id)

    def is_purchase_valid(self, account):
        """

        Test to see if this purchased asset can be accessed and is valid.

        :param account: account to used to check to see if this asset is purchased and has access using this account.
        :type account: :class:`.Account`

        :return: boolean value if this asset has been purchased
        :type: boolean
        """
        if not self.is_purchased:
            return False

        if not isinstance(account, Account):
            raise TypeError('You need to pass an Account object')

        if not account.is_valid:
            raise ValueError('You must pass a valid account')

        return self._agent.is_access_granted_for_asset(self._listing.asset, self._purchase_id, account)

    def invoke(self, account, payload):
        """

        Call invoke

        You can call the :func:`is_purchased` property before hand to check that you
        have already purchased this asset.

        :param account: account to used to consume this asset.
        :type account: :class:`.Account`
        :param str payload: Json payload that the invoke operation needs.

        :return: data returned from the asset , or False
        :type: object or False

        """
        if not self.is_purchased:
            return False

        if not isinstance(account, Account):
            raise TypeError('You need to pass an Account object')

        if not account.is_valid:
            raise ValueError('You must pass a valid account')
        logger.info(f'calling invoke in operation.py with payload: {payload}')
        return self._agent.invoke_operation(self._listing, self._purchase_id, account,
                                            payload)

    @property
    def get_type(self):
        return "operation"

    @property
    def is_purchased(self):
        """
        :return: True if this asset is a purchased asset.
        :type: boolean
        """
        return not self._purchase_id is None



"""koi client ."""
import sys,traceback
import json
import logging
import os

import requests
from tqdm import tqdm

from squid_py.agreements.service_agreement import ServiceAgreement
from squid_py.exceptions import OceanInitializeServiceAgreementError

from squid_py.brizo.brizo import Brizo
logger = logging.getLogger(__name__)


class KoiClient:
    """

    The main functions available are:
    - initialize_service_agreement
    - consume_service
    - run_compute_service (not implemented yet)

    """
    _http_client = requests

    @staticmethod
    def set_http_client(http_client):
        """Set the http client to something other than the default `requests`"""
        KoiClient._http_client = http_client

    @staticmethod
    def initialize_service_agreement(did, agreement_id, service_definition_id, signature,
                                     account_address,
                                     purchase_endpoint):
        """
        Send a request to the service provider (purchase_endpoint) to initialize the service
        agreement for the asset identified by `did`.

        :param did: str -- id of the asset includes the `did:op:` prefix
        :param agreement_id: hex str
        :param service_definition_id: str -- identifier of the service inside the asset DDO
        :param signature: hex str -- signed agreement hash
        :param account_address: hex str -- ethereum address of the consumer signing this agreement
        :param purchase_endpoint: str -- url of the service provider
        :return:
        """
        paymap={
            'asset-did': did,
            'service-agreement-id': agreement_id,
            'service-definition-id':service_definition_id,
            'signature': signature,
            'consumer-address': account_address
        }
        payload = json.dumps(paymap)
        logger.info(f' Koi client sending request to server {paymap}')
        response = KoiClient._http_client.post( purchase_endpoint, data=payload, headers={'content-type': 'application/json'})

        if response and hasattr(response, 'status_code'):
            if response.status_code != 201:
                msg = (f'Initialize service agreement failed at the purchaseEndpoint '
                       f'{purchase_endpoint}, reason {response.text}, status {response.status_code}'
                       )
                logger.error(msg)
                raise OceanInitializeServiceAgreementError(msg)

            logger.debug(
                f'Service agreement initialized successfully, service agreement id {agreement_id},'
                f' purchaseEndpoint {purchase_endpoint}')
            return True

    @staticmethod
    def invoke_service(service_agreement_id, service_endpoint, account_address, invoke_payload):
        """
        Call the Koi endpoint to get access to the different files that form the asset.

        :param service_agreement_id: Service Agreement Id, str
        :param service_endpoint: Url to consume, str
        :param account_address: ethereum address of the consumer signing this agreement, hex-str
        :param files: List containing the files to be consumed, list
        :param destination_folder: Path, str
        :return:
        """
        #TODO remove this hardcoding
        service_endpoint='http://localhost:8031/api/v1/brizo/services/invoke'
        logger.info(f'invoke endpoint with this url: {service_endpoint}')
        payload=json.dumps({'operationSAInfo':{'consumer-address': account_address, 'service-agreement-id': service_agreement_id}, 
            'payload':invoke_payload})
        response = KoiClient._http_client.post( service_endpoint, data=payload, headers={'content-type': 'application/json'}) 
            
        logger.info(f'invoke endpoint response : {response}')

        if response.status_code == 200: 
            logger.info(f'response successful') 
        else: 
            logger.warning(f'consume failed: {response.reason}')
        return json.loads(response.text)

    @staticmethod
    def get_brizo_url(config):
        """
        Return the Brizo component url.

        :param config: Config
        :return: Url, str
        """
        brizo_url = 'http://localhost:8031'
        if config.has_option('resources', 'brizo.url'):
            brizo_url = config.get('resources', 'brizo.url') or brizo_url

        brizo_path = '/api/v1/brizo'
        return f'{brizo_url}{brizo_path}'

    @staticmethod
    def get_purchase_endpoint(config):
        """
        Return the endpoint to purchase the asset.

        :param config:Config
        :return: Url, str
        """
        return f'{Brizo.get_brizo_url(config)}/services/access/initialize'

    @staticmethod
    def get_service_endpoint(config):
        """
        Return the url to consume the asset.

        :param config: Config
        :return: Url, str
        """
        return f'{Brizo.get_brizo_url(config)}/services/invoke'



"""


This class mocks the Surfer web service api.

I have been trying to get as close as the current surfer API, including the error
messages returned

see https://github.com/DEX-Company/surfer/blob/develop/src/main/clojure/surfer/handler.clj

"""

import pytest
import os
import logging
import json
import re

from urllib.parse import urlparse
from web3 import Web3

from unittest.mock import Mock

from starfish.logging import setup_logging

setup_logging(level=logging.DEBUG)
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("web3").setLevel(logging.WARNING)

Asset_storage = {}



class SurferMock(object):

    def __init__(self, url):
        self._url = url


    def put(self, url, data=None, headers=None):
        return self._route('put', url, data, headers)
    
    def get(self, url, data=None, headers=None):
        return self._route('get', url, data, headers)

    def post(self, url, json=None, headers=None):
        return self._route('post', url, json, headers)

    def metadata_api(self, method_type, path_items, data):
        if method_type == 'put':
            asset_id = ''
            if len(path_items) > 5:
                asset_id = path_items[5]
            metadata = data
            
            asset_hash = Web3.toHex(Web3.sha3(metadata.encode()))[2:]
            if asset_hash == asset_id:
                # now save in memory the asset metadata for the read? ( TODO: read)
                Asset_storage[asset_id] = metadata
                return SurferMock._response(200, asset_hash)
            return SurferMock._response(400, f'Invalid ID for metadata, expected: "{asset_hash}" got "{asset_id}"')
        elif method_type == 'get':            
            asset_id = ''
            if len(path_items) > 5:
                asset_id = path_items[5]
            if Asset_storage[asset_id]:
                return SurferMock._response(200, Asset_storage[asset_id])
            return SurferMock._response(404, 'Asset id not found')
        return SurferMock._response(400, 'Bad request')

    def _route(self, method_type, url, data=None, headers=None):
        url_split = urlparse(url)
        assert(url_split)
        if url_split.path:
            if not re.match('^/api/v1', url_split.path):
                return SurferMock._repsonse(404, 'path not found')
            path_items = url_split.path.split('/')
            if len(path_items) < 4:
                return SurferMock._repsonse(404, 'path not found')

            if path_items[3] == 'meta':
                if path_items[4] == 'data':
                    return self.metadata_api(method_type, path_items, data)
                else:
                    return SurferMock._repsonse(404, 'path not found')
            else:
                return SurferMock._repsonse(404, 'path not found')
                                
        return SurferMock._response(400, 'Bad request')
        

    @staticmethod
    def _response(status=200, content="CONTENT", json_data=None):
        response = Mock()
        response.status_code = status
        response.content = content
        # add json data
        if json_data:
            response.json = Mock(return_value=json_data)
        return response



"""

Agent class for Invokable Ocean Agents

"""

import logging
from starfish.operation.operation import Operation

logger = logging.getLogger('starfish.invoke_agent')

class InvokeAgent:
    """

    Invoke Agent class allows get a list of operations and invoke them.

    """
    def __init__(self, **kwargs):
        """init a standard ocean object"""

        self._koi_url = kwargs.get('koi_url', 'http://localhost:8031/api/v1/brizo/services/')

    def get_operations(self):
        """
        returns a list of operations and their associated DIDs.
        """
        return []
#        r=requests.get(self._koi_url+'operations')
#        j=json.loads(r.text)
#        self.operations=dict(zip([i['name'] for i in j],[i['did'] for i in j]))
#        return self.operations

    def get_operation(self, did):
        return Operation(self, did)

    def get_listing(self, listing_id):
        """

        Return an listing from the given listing_id.

        :param str listing_id: Id of the listing.

        :return: a registered listing given a Id of the listing
        :type: :class:`.Listing` class

        """

    def purchase(self, listing, account):
        """

        Purchase an asset using it's listing and an account.-will be supported in future

        """
        pass



"""

CryptoHash utils

"""


from Crypto.Hash import SHA3_256
from web3 import Web3

def hash_sha3_256(data):
    if isinstance(data, str):
        data = data.encode('utf-8')

    messageDigest = SHA3_256.new()
    messageDigest.update(data)
    return messageDigest.hexdigest()


def hash_keccak_256(data):
    if isinstance(data, str):
        data = data.encode('utf-8')

    return Web3.toHex(Web3.sha3(data))[2:]



import os
import os.path
import traceback
import re
from starfish.command_line.console_output import ConsoleOutput
from starfish.command_line.json_output import JSONOutput
from starfish.exceptions import OceanCommandLineError
from starfish import Ocean


HEX_STRING_FORMAT = '[0-fx]+'
INTEGER_FORMAT = '[0-9]+'
URL_STRING_FORMAT = '^http.*'


def get_default(args, index, default=None):
    if len(args) > index:
        return args[index]
    return default

def assert_arg_match(args, index, patterns, message):
    value = get_default(args, index)
    if value:
        for pattern in patterns:
            if re.match(pattern, value):
                return value
    raise OceanCommandLineError(message)

def assert_arg_filename(args, index, message):
    value = get_default(args, index)
    if value and os.path.exists(value):
        return value
    raise OceanCommandLineError(message)


class CommandLine:
    def __init__(self, **kwargs):
        self._ocean = Ocean(**kwargs)
        self._output = kwargs.get('output', None)
        if self._output is None:
            if kwargs.get('is_json', False):
                self._output = JSONOutput()
            else:
                self._output = ConsoleOutput()

    def call(self, command, args):
        try:
            command_func = getattr(self, command.replace('-', '_'))
            command_func(args)
            return self._output
        except Exception as exception:
            if isinstance(exception, AttributeError):
                raise OceanCommandLineError(f'Cannot find command "{command}"')
            print(exception)
            traceback.print_exc()
            raise OceanCommandLineError(f'Command failed "{command}"')

    def balance(self, args):
        self._output.clear()
        filter_account_id = None
        if args:
            filter_account_id = assert_arg_match(args, 0,
                                                 [HEX_STRING_FORMAT,
                                                  INTEGER_FORMAT],
                                                 'You must provide a valid account id')
        self._output.set_header('index', 'Index', '{:10}')
        self._output.set_header('account', 'Account', '{:44}')
        self._output.set_header('tokens', 'Ocean Tokens', '{:>20}')
        self._output.set_header('ether', 'Ether', '{:>40}')
        self._output.show_header()
        index = 0
        for account_id, account in self._ocean.accounts.items():
            if not filter_account_id or account_id == filter_account_id:
                self._output.set_item('index', str(index))
                self._output.set_item('account', account_id)
                self._output.set_item('tokens', account.ocean_balance)
                self._output.set_item('ether', account.ether_balance)
                self._output.show_items()
            index = index + 1




import logging
import json
from base64 import b64encode
from Cryptodome.Hash import SHA256
from Cryptodome.PublicKey import RSA
from Cryptodome.Signature import PKCS1_v1_5

from starfish.ddo.authentication import Authentication


from ocean_utils.ddo.ddo import DDO

from ocean_utils.ddo.public_key_base import PUBLIC_KEY_STORE_TYPE_PEM, PublicKeyBase
from ocean_utils.ddo.constants import KEY_PAIR_MODULUS_BIT, DID_DDO_CONTEXT_URL, PROOF_TYPE
from ocean_utils.ddo.public_key_rsa import PublicKeyRSA, PUBLIC_KEY_TYPE_RSA
from ocean_utils.ddo.service import Service


logger = logging.getLogger('starfish.ddo')

class StarfishDDO(DDO):

    def set_service_endpoint(self, service_type, value):
        """

        Set the default service endpoint with a new value

        :param service_type: Type of service to set
        :param value: value to set as the new value

        """
        service = self.get_service(service_type)
        if service:
            service._service_endpoint = value

    def add_authentication(self, key_id, authentication_type=None):
        """
        Add a authentication public key id and type to the list of authentications.
        :param key_id: Key id, Authentication
        :param authentication_type: Authentication type, str
        :return: None
        """
        if isinstance(key_id, Authentication):
            # adding an authentication object
            authentication = key_id
        elif isinstance(key_id, PublicKeyBase):
            public_key = key_id
            # this is going to be a embedded public key
            authentication = Authentication(public_key, public_key.get_authentication_type())
        else:
            # with key_id as a string, we also need to provide the authentication type
            if authentication_type is None:
                raise ValueError
            authentication = Authentication(key_id, authentication_type)

        logger.debug(f'Adding authentication {authentication}')
        self._authentications.append(authentication)

    def add_signature(self, public_key_store_type=PUBLIC_KEY_STORE_TYPE_PEM, is_embedded=False):
        """
        Add signature.
        Add a signature with a public key and authentication entry for validating this DDO
        returns the private key as part of the private/public key pair.
        :param public_key_store_type: Public key store type, str
        :param is_embedded: bool
        :return Private key pem, str
        """

        key_pair = RSA.generate(KEY_PAIR_MODULUS_BIT, e=65537)
        public_key_raw = key_pair.publickey()
        private_key_pem = key_pair.exportKey("PEM")

        # find the current public key count
        next_index = self._get_public_key_count() + 1
        key_id = f'{self._did}#keys={next_index}'

        public_key = PublicKeyRSA(key_id, owner=key_id)

        public_key.set_encode_key_value(public_key_raw, public_key_store_type)

        if is_embedded:
            # also add the authentication key as embedded key with the authentication
            self.add_authentication(public_key)
        else:
            # add the public key to the DDO list of public keys
            self.add_public_key(None, public_key)

            # add the public key id and type for this key to the authentication
            self.add_authentication(public_key.get_id(), public_key.get_authentication_type())

        logger.debug('Adding signature to the ddo object.')
        return private_key_pem

    def add_public_key(self, did, public_key):
        """
        Add a public key object to the list of public keys.
        :param public_key: Public key, PublicKeyHex
        """
        if did is None:
            self._public_keys.append(public_key)
        else:
            logger.debug(f'Adding public key {public_key} to the did {did}')
            self._public_keys.append(
                PublicKeyBase(did, **{"owner": public_key, "type": "EthereumECDSAKey"}))

    def add_proof_keeper(self, text, publisher_account, keeper):
        """Add a proof to the DDO, based on the public_key id/index and signed with the private key
        add a static proof to the DDO, based on one of the public keys."""

        # just incase clear out the current static proof property
        self._proof = None
        self._proof = {
            'type': PROOF_TYPE,
            'created': DDO._get_timestamp(),
            'creator': publisher_account.address,
            'signatureValue': keeper.sign_hash(text, publisher_account),
        }

    def add_proof(self, authorisation_index, private_key=None, signature_text=None):
        """Add a proof to the DDO, based on the public_key id/index and signed with the private key
        add a static proof to the DDO, based on one of the public keys."""

        # find the key using an index, or key name
        if isinstance(authorisation_index, dict):
            self._proof = authorisation_index
            return

        if private_key is None:
            raise ValueError

        authentication = self._authentications[authorisation_index]
        if not authentication:
            raise IndexError

        if authentication.is_public_key():
            sign_key = authentication.get_public_key()
        else:
            sign_key = self.get_public_key(authentication.get_public_key_id())

        if sign_key is None:
            raise IndexError

        # get the signature text if not provided

        if signature_text is None:
            hash_text_list = self._hash_text_list()
            signature_text = "".join(hash_text_list)

        # just incase clear out the current static proof property
        self._proof = None

        signature = StarfishDDO.sign_text(signature_text, private_key, sign_key.get_type())

        self._proof = {
            'type': sign_key.get_type(),
            'created': DDO._get_timestamp(),
            'creator': sign_key.get_id(),
            'signatureValue': b64encode(signature).decode('utf-8'),
        }

    def as_dictionary(self, is_proof=True):
        """
        Return the DDO as a JSON dict.
        :param if is_proof: if False then do not include the 'proof' element.
        :return: dict
        """
        if self._created is None:
            self._created = DDO._get_timestamp()

        data = {
            '@context': DID_DDO_CONTEXT_URL,
            'id': self._did,
            'created': self._created,
        }
        if self._public_keys:
            values = []
            for public_key in self._public_keys:
                values.append(public_key.as_dictionary())
            data['publicKey'] = values
        if self._authentications:
            values = []
            for authentication in self._authentications:
                values.append(authentication.as_dictionary())
            data['authentication'] = values
        if self._services:
            values = []
            for service in self._services:
                values.append(service.as_dictionary())
            data['service'] = values
        if self._proof and is_proof:
            data['proof'] = self._proof

        return data

    def _read_dict(self, dictionary):
        """Import a JSON dict into this DDO."""
        values = dictionary
        self._did = values['id']
        self._created = values.get('created', None)
        if 'publicKey' in values:
            self._public_keys = []
            for value in values['publicKey']:
                if isinstance(value, str):
                    value = json.loads(value)
                self._public_keys.append(DDO.create_public_key_from_json(value))
        if 'authentication' in values:
            self._authentications = []
            for value in values['authentication']:
                if isinstance(value, str):
                    value = json.loads(value)
                self._authentications.append(StarfishDDO.create_authentication_from_json(value))
        if 'service' in values:
            self._services = []
            for value in values['service']:
                if isinstance(value, str):
                    value = json.loads(value)
                service = Service.from_json(value)
                service.set_did(self._did)
                self._services.append(service)
        if 'proof' in values:
            self._proof = values['proof']

    def _hash_text_list(self):
        """Return a list of all of the hash text."""
        hash_text = []
        if self._created:
            hash_text.append(self._created)

        if self._public_keys:
            for public_key in self._public_keys:
                if public_key.get_type():
                    hash_text.append(public_key.get_type())
                if public_key.get_value():
                    hash_text.append(public_key.get_value())

        if self._authentications:
            for authentication in self._authentications:
                if authentication.is_public_key():
                    public_key = authentication.get_public_key()
                    if public_key.get_type():
                        hash_text.append(public_key.get_type())
                    if public_key.get_value():
                        hash_text.append(public_key.get_value())

        if self._services:
            for service in self._services:
                hash_text.append(service.type)
                hash_text.extend(service.endpoints[0])

        # if no data can be found to hash then raise an error
        if not hash_text:
            raise ValueError
        return hash_text

    @staticmethod
    def create_authentication_from_json(values):
        """Create authentitaciton object from a JSON string."""
        key_id = values.get('publicKey')
        authentication_type = values.get('type')
        if not key_id:
            raise ValueError(
                f'Invalid authentication definition, "publicKey" is missing: {values}')
        if isinstance(key_id, dict):
            public_key = DDO.create_public_key_from_json(key_id)
            authentication = Authentication(public_key, public_key.get_authentication_type())
        else:
            authentication = Authentication(key_id, authentication_type)

        return authentication

    @staticmethod
    def sign_text(text, private_key, sign_type=PUBLIC_KEY_TYPE_RSA):
        """Sign some text using the private key provided."""
        if sign_type == PUBLIC_KEY_TYPE_RSA:
            signer = PKCS1_v1_5.new(RSA.import_key(private_key))
            text_hash = SHA256.new(text.encode('utf-8'))
            signed_text = signer.sign(text_hash)
            return signed_text

        raise NotImplementedError



import pytest
import pathlib
import json
import datetime
import logging
from unittest.mock import Mock

RESOURCES_PATH = pathlib.Path.cwd() / 'tests' / 'resources'
METADATA_SAMPLE_FILE = RESOURCES_PATH / 'metadata' / 'sample_metadata1.json'
TEST_ASSET_FILE = RESOURCES_PATH / 'test_asset_file.txt'
TEST_ASSET_REMOTE = 'https://oceanprotocol.com/tech-whitepaper.pdf'


# set debug logging
logging.basicConfig(level=logging.DEBUG)


TEST_LISTING_DATA = {
    'name': 'Test file asset',
    'description': 'Test asset for sale',
    'dateCreated': datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ'),
    'author': 'Test starfish',
    'license': 'Closed',
    'price': 3.141592,              # price is in ocean tokens
    'extra_data': 'Some extra data',
    'tags': ['asset', 'sale', 'test', 'starfish'],
}


@pytest.fixture(scope="module")
def metadata():
    # load in the sample metadata
    assert METADATA_SAMPLE_FILE.exists(), "{} does not exist!".format(METADATA_SAMPLE_FILE)
    metadata = None
    with open(METADATA_SAMPLE_FILE, 'r') as file_handle:
        metadata = json.load(file_handle)

    return metadata

@pytest.fixture(scope="module")
def resources():
    data = Mock()
    data.asset_file = TEST_ASSET_FILE
    data.metadata = METADATA_SAMPLE_FILE
    data.asset_remote = TEST_ASSET_REMOTE
    data.listing_data = TEST_LISTING_DATA
    return data





"""

    Command line JSON output

"""
import json

from .base_output import BaseOutput

class JSONOutput(BaseOutput):


    def show(self, text=None, name=None):
        if self._rows:
            print(json.dumps(self._rows))

    def show_item(self, name, value, format_text=None):
        item = {name: value}
        print(json.dumps(item))

    def show_header(self):
        pass

    def set_item(self, name, value):
        self._items[name] = value

    def show_items(self):
        self._rows.append(self._items)
        self._items = {}

    def flush(self):
        if self._rows:
            print(json.dumps(self._rows))
        if self._items:
            print(json.dumps(self._items))
        self._rows = []
        self._items = {}



"""

Account class to provide basic functionality for all Ocean Accounts

"""

from web3 import Web3

class Account():
    """

    Account class, adds functionality for an account to be used by the Ocean network.

    At the moment the Account object is created by the :class:`Ocean` class.

    :param ocean: Ocean object
    :type ocean: :class:`.Ocean`
    :param address: address or dict of the account details
    :type address: string or dict
    :param password: password for the account
    :type password: str or None

    If the address parameter is a string then it's the account address.
    If dict then the dict can be the following format: ::

        {
            'address': 'xxxx',
            'password': 'yyyy',
        }

    """

    def __init__(self, ocean, address, password=None, keyfile=None, agent_adapter=None):
        """init a standard ocean agent"""
        self._ocean = ocean
        self._address = None
        self._password = None
        self._agent_adapter = agent_adapter
        if self._agent_adapter is None:
            self._agent_adapter = self._ocean.get_squid_agent_adapter()


        if isinstance(address, dict):
            self._address = address.get('address')
            self._password = address.get('password')
            self._keyfile = address.get('keyfile')
        elif isinstance(address, (tuple, list)):
            self._address = address[0]
            if len(address) > 1:
                self._password = address[1]
            if len(address) > 2:
                self._keyfile = address[2]
        elif isinstance(address, str):
            self._address = address
            self._password = password
            self._keyfile = keyfile


    def request_tokens(self, amount):
        """

        For **Testing Only**

        Request some ocean tokens to be transfere to this account address

        :param number amount: The amount of ocean tokens to transfer ( *Money for nothing* )

        :return: number of tokens transfered
        :type: int

        >>> account.request_tokens(100)
        100
        """
        adapter = self.agent_adapter
        if adapter:
            return adapter.request_tokens(amount, self.agent_adapter_account)
        return 0

    def approve_tokens(self, spender_address, amount):
        """
        Approve some ocean tokens to be transfere from this account address by another Spender address

        :param number amount: The amount of ocean tokens to approve
        :param address spender_address: The address of spender
        :return: boolean
        """
        adapter = self.agent_adapter
        if adapter:
            amount = Web3.toWei(amount, 'ether')
            return adapter.approve_tokens(spender_address, amount, self)
        return False

    def transfer_ether(self, to_account, amount_ether):
        """

        Transfer ether from this account to another account

        :param to_account: To account object or account address
        :type to_account: str for an account address or :class:'.Account` object

        ;param amount_ether: amount in ether to transfer
        :return: number of ether transfered

        """

        amount_wei = Web3.toWei(amount_ether, 'ether')
        to_address = to_account
        if isinstance(to_account, Account):
            to_address = to_account.address

        adapter = self.agent_adapter
        if adapter:
            return adapter.transfer_ether(self.agent_adapter_account, to_address, amount_wei)
        return 0

    def transfer_token(self, to_account, amount_token):
        """

        Transfer ocean tokens from this account to another account

        :param to_account: To account object or account address
        :type to_account: str for an account address or :class:'.Account` object

        ;param amount_ether: amount in ocean tokens to transfer
        :return: number of tokens transfered

        """

        amount_vodka = Web3.toWei(amount_token, 'ether')
        to_address = to_account
        if isinstance(to_account, Account):
            to_address = to_account.address

        adapter = self.agent_adapter
        if adapter:
            return adapter.transfer_tokens(self.agent_adapter_account, to_address, amount_vodka)
        return 0

    def is_address_equal(self, address):
        """

        Compares two addresses if are equal. Both addresses are converted to checksum
        address and then compared.

        :param str address: address to compare with this object's address
        :return: True if the param address is the same is the one held in this account
        :type: boolean

        >>> account = ocean.get_account('0x00bd138abd70e2f00903268f3db08f2d25677c9e')
        >>> account.is_address_equal('0x00Bd138aBD70e2F00903268F3Db08f2D25677C9e')
        True
        """
        return self.as_checksum_address == Web3.toChecksumAddress(address)

    @property
    def ocean(self):
        """

        Return the :class:`.Ocean` object for this account
        :return: The main ocean class
        :type: :class:`.Ocean`

        """
        return self._ocean


    @property
    def is_password(self):
        """
        Return True if the password has been set, else return False

        """
        return self._password is not None

    @property
    def address(self):
        """

        Return the account address for this account

        :return: address
        :type: str

        >>> account = ocean.get_account('0x00bd138abd70e2f00903268f3db08f2d25677c9e')
        >>> account.address
        0x00bd138abd70e2f00903268f3db08f2d25677c9e
        """
        return self._address

    @property
    def as_checksum_address(self):
        """

        Return the address as a checksum address

        :return: checksum address
        :type: str

        >>> account = ocean.get_account('0x00bd138abd70e2f00903268f3db08f2d25677c9e')
        >>> account.as_checksum_address
        0x00Bd138aBD70e2F00903268F3Db08f2D25677C9e

        """
        if self._address:
            return Web3.toChecksumAddress(self._address.lower())
        return None

    @property
    def password(self):
        """

        Return the account password for this account

        :return: password
        :type: str

        >>> account.password
        secret

        """
        return self._password

    def set_password(self, password):
        """

        Set the password for this account

        :param str password: Password to set for this account

        >>> account.set_password('new secret')
        >>> account.password
        new secret
        """
        self._password = password

    @property
    def keyfile(self):
        return self._keyfile

    @property
    def key_file(self):
        """
        Compatability field, so that this account behaves the same as a squid account
        """
        return self._keyfile

    @property
    def is_valid(self):
        return self._address and self._password and self._keyfile

    @property
    def ocean_balance(self):
        """

        Get the number of ocean tokens

        :return: number of ocean tokens
        :type: number

        >>> account.ocean_balance
        101

        """
        adapter = self.agent_adapter
        if adapter:
            balance = adapter.get_account_balance(self.agent_adapter_account)
            if balance:
                return Web3.fromWei(balance.ocn, 'ether')
        return 0

    @property
    def ocean_balance_raw(self):
        """

        Get the number of ocean tokens

        :return: number of ocean tokens
        :type: number

        >>> account.ocean_balance_raw
        101

        """
        adapter = self.agent_adapter
        if adapter:
            balance = adapter.get_account_balance(self.agent_adapter_account)
            if balance:
                return balance.ocn
        return 0

    @property
    def ether_balance(self):
        """

        Get the number of ocean tokens

        :return: number of ocean tokens
        :type: number

        >>> account.ether_balance
        1000000001867769600000000000

        """
        adapter = self.agent_adapter
        if adapter:
            balance = adapter.get_account_balance(self.agent_adapter_account)
            if balance:
                return Web3.fromWei(balance.eth, 'ether')
        return 0

    @property
    def agent_adapter(self):
        return self._agent_adapter

    @agent_adapter.setter
    def agent_adapter(self, agent_adapter):
        self._agent_adapter = agent_adapter

    @property
    def agent_adapter_account(self):
        """

        Return the squid account object used for squid services, curretly only hosted accounts can be used
        :params obj adapter: middleware agent adapter to get the actual account object for this adapter
        :return: The squid account object
        :type: object

        """

        return self._agent_adapter.get_account(self.as_checksum_address, self._password, self._keyfile)

    def __str__(self):
        return f'Account: {self.address}'



"""
    Inovke Asset
"""

from starfish.asset.asset_base import AssetBase

class OperationAsset(AssetBase):

    """

    Operation asset can be used to perform a invokable service

    :param metadata: Optional dictionary metadata to provide for the asset
        if non used then the class will generate a default metadata based on the data provided
    :type metadata: None or dict
    :param did: Optional did of the asset if it's registered
    :type did: None or str

    """
    def __init__(self, metadata=None, did=None, metadata_text=None):
        if not isinstance(metadata, dict):
            raise ValueError('metadata must be a dict')

        AssetBase.__init__(self, metadata, did, metadata_text)

    def create(name, metadata=None, did=None):
        """

        Create a new OperationAsset.

        :param str name: Name of the asset
        :param dict metadata: Optional metadata to add to the assets metadata
        :param str did: Option DID to assign to this asset

        """
        metadata = AssetBase.generateMetadata(name, 'operation', metadata)
        return OperationAsset(metadata, did)

    def is_mode(self, mode_type):
        """

        Check to see if this operation supports the mode provided.

        :param str mode_type: Mode type to check to see if this operation supports
        :return: Return True if this mode is supported
        :type: boolean

        """
        try:
            return mode_type in self._metadata['operation']['modes']
        except:
            raise ValueError('Metadata does not contain operation->modes structure')




import pytest


from starfish.agent.services import (
    Services,
    SERVICES,
    ALL_SERVICES
)


def test_init():
    services = Services('http://localhost')
    services.add('metadata')
    result = services.as_dict
    assert(result)
    assert('metadata' in result)
    assert(result['metadata']['url'] == 'http://localhost/api/v1/meta/data')


    services = Services('http://localhost', version='v99')
    services.add('metadata')
    result = services.as_dict
    assert(result)
    assert('metadata' in result)
    assert(result['metadata']['url'] == 'http://localhost/api/v99/meta/data')
    assert('metadata' in services.names)


    services = Services('http://localhost', service_list=['metadata', 'storage'])
    result = services.as_dict
    assert(result)
    assert('metadata' in result)
    assert(result['metadata']['url'] == 'http://localhost/api/v1/meta/data')
    assert('storage' in result)
    assert(result['storage']['url'] == 'http://localhost/api/v1/assets')


    services = Services('http://localhost', service_list=ALL_SERVICES)
    result = services.as_dict
    assert(result)
    assert(services.names == ALL_SERVICES)

    with pytest.raises(ValueError):
        services.add('bad-name')



"""
    Bundle Asset
"""

from starfish.asset.asset_base import AssetBase

class BundleAsset(AssetBase):
    """

    Bundle asset can be used to hold many assets

    :param metadata: Dictionary metadata to provide for the asset
    :type metadata: dict
    :param did: Optional did of the asset if it's registered
    :type did: None or str

    """
    def __init__(self, metadata, did=None, metadata_text=None):

        if not isinstance(metadata, dict):
            raise ValueError('metadata must be a dict')
        AssetBase.__init__(self, metadata, did, metadata_text)
        self._assets = {}

    @staticmethod
    def create(name, metadata=None, did=None):
        """

        Create a new empty BundleAsset

        :param str name: Name of the asset to create
        :param dict metadata: Optional metadata to add to the assets metadata
        :param str did: Option DID to assign to this asset

        :return: a new BundleAsset
        :type: :class:`.BundleAsset`
        """

        metadata = AssetBase.generateMetadata(name, 'bundle', metadata)
        return BundleAsset(metadata, did)

    def add(self, name, asset):
        """

        Add an asset to the bundle. Any asset object

        :param string name: name of the asset in the bundle
        :param asset: asset to add to the bundle collection
        :type asset: :class:`.Asset` object

        """

        if not isinstance(name, str):
            raise TypeError('You need to pass the asset name as a string')
        if not isinstance(asset, AssetBase):
            raise TypeError('You need to pass an Asset object')

        self._assets[name] = asset

    def get_asset(self, name):
        """

        Return the asset in the bundle based on it's index number

        :param str name: name of the asset in this bundle

        :return: asset is returned
        :type: :class:`.Asset` object

        :raises: ValueError if name is not found in the bundle
        """
        if isinstance(name, str):
            if name not in self._assets:
                raise ValueError(f'Cannot find asset named {name}')
        if isinstance(name, int):
            if name < 0 or name > self.asset_count:
                raise ValueError(f'Cannot find asset at index {name}')
            name = self.asset_names[name]
        return self._assets[name]

    def asset_remove(self, name):
        """

        Remove an asset from the bundle with a given name

        :param str name: name of the asset to remove
        :return: the removed asset
        :type: :class:`.Asset` object

        """
        if self._assets is None or name not in self._assets:
            raise ValueError(f'Cannot find asset named {name}')
        asset = self._assets[name]
        del self._assets[name]
        return asset


    def __iter__(self):
        """

        Iterator method. This allows you to do the following::

            for index, asset in bundle:
                print('my assets in the bundle are', index, asset)
        """
        self._iter_index = 0
        return self

    def __next__(self):
        """

        Provide the __next__ method of a iterator.

        :return: index, asset
        :type: int, :class:.`Asset` object

        """
        if self._iter_index < self.asset_count:
            index = self._iter_index
            name = self.asset_names[index]
            asset = self._assets[name]
            self._iter_index += 1
            return name, asset
        else:
            raise StopIteration

    def __getitem__(self, name):
        return self._assets[name]

    def get_asset_at_index(self, index):
        """ return the asset based on the index of available assets """
        return list(self._assets.values())[index]

    @property
    def asset_count(self):
        """

        Return the number of assets in this bundle

        :return: count of assets
        :type: int
        """
        return len(self._assets.keys())

    @property
    def asset_items(self):
        return self._assets.items()

    @property
    def asset_names(self):
        return list(self._assets.keys())

    @property
    def is_bundle(self):
        """

        Return True if this asset is a bundle asset and can contain sub assets ( Asset Bundle )

        :return: True if sub assets can be held within this asset
        :type: boolean

        """
        return True



"""
    test_13_asset_listing_search

    As a developer building client code to interact with an Ocean marketplace,
    I need a way to search available asset listings

"""

import secrets

from starfish.asset import DataAsset


def test_13_asset_listing_search(resources, remote_agent):
    test_data = secrets.token_hex(1024)
    asset = DataAsset.create('TestAsset', test_data)
    listing = remote_agent.register_asset(asset, resources.listing_data)
    assert(listing)
    assert(listing.asset)



#!/usr/bin/env python
"""

    Tests for `ocean agent command line`.

"""
import unittest
import random
import os
import logging

from pytest import (
    mark,
    raises,
)

from starfish.command_line import CommandLine
from starfish.command_line.pytest_output import PyTestOutput


def test_command_line_balance():

    test_output = PyTestOutput()
    command_line = CommandLine(output = test_output)
    args = []
    output = command_line.call('balance', args)
    index = 0
    for row in output.rows:
        assert int(row['index']) == index
        assert row['account'] == command_line._ocean.accounts[row['account']].address
        assert 'tokens' in row
        assert 'ether' in row
        index += 1

"""
def test_command_line_asset():

    logging.basicConfig(level = logging.DEBUG)
    logger = logging.getLogger(os.path.basename(__file__))

    command_line = CommandLine(ethereum_url = ETHEREUM_URL,
        keeper_contracts_path = KEEPER_CONTRACT_FILES,
        meta_storage_service = META_STORAGE_URL,
        asset_provider_service = ASSET_PROVIDER_URL,
        logger = logger,
        output = PyTestOutput()
    )
    ocean = command_line.open()

    # register an asset
    test_name = 'Name for asset ' + ASSET_FILE
    args = [str(ASSET_OWNER_ACCOUNT_INDEX), ASSET_FILE, test_name]
    output = command_line.call('register', args)
    result = output.get_items()
    assert result != None
    assert result['owner'] == ocean.accounts[ASSET_OWNER_ACCOUNT_INDEX]
    asset_id = result['asset id']

    # get asset info
    args = [asset_id]
    output = command_line.call('info', args)
    result = output.get_items()
    assert result != None
    assert result['owner:'] == ocean.accounts[ASSET_OWNER_ACCOUNT_INDEX]
    assert result['asset_id'] == asset_id

    # get asset list
    args = []
    output = command_line.call('list', args)
    result = output.get_rows()
    assert result != None
    is_found = False
    for row in result:
        if row['asset_id'] == asset_id:
            is_found = True
            break
    assert is_found == True

    # upload an asset
    args = [asset_id, str(ASSET_PRICE), ASSET_FILE]
    output = command_line.call('upload', args)
    result = output.get_items()
    assert result != None
    assert result['asset_id'] == asset_id
    assert result['price'] == ASSET_PRICE

    # asset list for sale

    args = ['sale']
    output = command_line.call('list', args)
    result = output.get_rows()
    assert result != None
    is_found = False
    for row in result:
        if row['asset_id'] == asset_id:
            is_found = True
            break
    assert is_found == True

    # credit test account with tokens
    args = [str(ASSET_BUYER_ACCOUNT_INDEX), str(ASSET_PRICE)]
    output = command_line.call('credit', args)
    result = output.get_items()
    assert result != None
    assert result['credit_amount'] == ASSET_PRICE
    assert result['owner'] == ocean.accounts[ASSET_BUYER_ACCOUNT_INDEX]

    # buy an asset
    args = [asset_id, str(ASSET_BUYER_ACCOUNT_INDEX)]
    output = command_line.call('buy', args)
    result = output.get_items()
    assert result != None
    assert result['access_id:'] != None
    assert result['secret token:'] != None
    assert result['from service id:'] != None

    access_id = result['access_id:']
    token = result['secret token:']


    # get asset list as sold
    args = ['sold', str(ASSET_BUYER_ACCOUNT_INDEX)]
    output = command_line.call('list', args)
    result = output.get_rows()
    assert result != None
    is_found = False
    for row in result:
        if row['access_id'] == access_id and row['asset_id'] == asset_id:
            is_found = True
    assert is_found == True

    #get asset data
    args = [access_id, token]
    output = command_line.call('get', args)
    result = output.get_items()
    assert result != None
    assert 'data' in result
    assert result['data'] != None
"""



"""
    test_05_agent_endpoint_update


    As a developer managing a Ocean Agent,
    I need to be able to update service endpoints for my Agent

"""

import pytest
import re

from starfish.agent import RemoteAgent
from starfish.ddo.starfish_ddo import StarfishDDO
from starfish.agent.services import Services, ALL_SERVICES



def test_05_agent_endpoint_update(ocean, config, remote_agent):
    endpoint = remote_agent.get_endpoint('DEP.Meta.v1')
    assert(endpoint)
    assert(re.search('meta/data', endpoint))

    services = Services(config.remote_agent_url, ['metadata'], 'v99')

    ddo = RemoteAgent.generate_ddo(services)
    new_endpoint_uri = '/app/v99/meta/test'
    new_agent = RemoteAgent(ocean, ddo=ddo)
    with pytest.raises(ValueError):
        new_endpoint = new_agent.get_endpoint('DEP.Meta.v99')



"""
    test_operation_asset

"""

import secrets
import logging
import json

from starfish.asset import OperationAsset

OPERATION_METADATA = {
    'name': 'operation',
    'type': 'operation',
    'operation': { 
        'modes': ['sync', 'async', 'test'],
    },
}


def test_init(metadata):
    asset = OperationAsset(OPERATION_METADATA)
    assert(asset)
    assert(isinstance(asset, OperationAsset))


def test_operation_asset_mode(ocean, metadata, config):
    
    asset = OperationAsset(OPERATION_METADATA)
    for mode_name in OPERATION_METADATA['operation']['modes']:
        assert(asset.is_mode(mode_name))
        assert(not asset.is_mode(f'{mode_name}-badmode'))




#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0


import logging
import os
import time
from datetime import datetime
from threading import Thread


from ocean_events_handler.event_handlers import accessSecretStore, lockRewardCondition
from ocean_utils.agreements.service_agreement import ServiceAgreement
from ocean_utils.agreements.service_agreement_template import ServiceAgreementTemplate
from ocean_utils.agreements.service_types import ServiceTypes
from ocean_utils.agreements.utils import get_sla_template
from ocean_events_handler.agreement_store.agreements import AgreementsStorage
from ocean_utils.did import id_to_did
from ocean_utils.did_resolver.did_resolver import DIDResolver
from ocean_keeper.web3_provider import Web3Provider

logger = logging.getLogger(__name__)


class StarfishAgreementRecord:

    def __init__(self, agreement_id, publisher, consumer, provider, did, condition_to_id):
        self.agreement_id = agreement_id
        self.publisher = publisher
        self.consumer = consumer
        self.provider = provider
        self.did = did
        self.condition_to_id = condition_to_id
        self.cond_state = {cond: 0 for cond in condition_to_id.keys()}
        self.created = int(datetime.now().timestamp())

    @property
    def completed(self):
        return len([1 for s in self.cond_state.values() if s >= 2]) == 3

    @property
    def missing_conditions(self):
        return {cond for cond, state in self.cond_state.items() if state < 2}

    def get_condition_id(self, cond):
        return self.condition_to_id.get(cond)

    def update_cond_state(self, cond, state):
        self.cond_state[cond] = state

    @property
    def report(self):
        return f'{self.agreement_id} -> {self.cond_state.items()}'


class StarfishEventsManager:
    """


    Manage the main keeper events listeners necessary for processing service agreements.

    on init
        if not db or not db schema -> create db and schema
        determine LAST_N_BLOCKS (allow setting from outside using an env var)
        read agreements from local database since LAST_N_BLOCKS
            keep set of completed/fulfilled (all conditions fulfilled) agreements to avoid reprocessing during event processing
            process events for unfulfilled agreements

    in watcher loop
        get AgreementCreated events since LAST_N_BLOCKS or LAST_PROCESSED_BLOCK whichever is larger
            try to overlap LAST_PROCESSED_BLOCK when grabbing events so we don't miss any events

    on new agreement (AgreementCreated event)
        save agreement to db
        init agreement conditions with unfulfilled status
        watch condition events
            on each condition event
                update agreement condition status


    """

    _instance = None

    EVENT_WAIT_TIMEOUT = 3600
    LAST_N_BLOCKS = 200

    def __init__(self, keeper, storage_path, account):
        self._keeper = keeper
        self._storage_path = storage_path
        self._account = account
        self._web3 = Web3Provider.get_web3()
        self._db = AgreementsStorage(self._storage_path)
        self._is_valid_callback = None

        self.known_agreement_ids = set()
        self.completed_ids = set()
        self.other_agreement_ids = set()

        # prepare condition names and events arguments dict
        sla_template = ServiceAgreementTemplate(template_json=get_sla_template())
        self.condition_names = [cond.name for cond in sla_template.conditions]
        self.event_to_arg_names = sla_template.get_event_to_args_map(self._keeper.contract_name_to_instance)

        db = self.db
        db.create_tables()
        # get largest block_number from db or `latest` if db has no data
        self.last_n_blocks = os.getenv('OCN_EVENTS_MONITOR_LAST_N_BLOCKS', self.LAST_N_BLOCKS)
        self.latest_block = self._web3.eth.blockNumber
        db_latest = db.get_latest_block_number() or self.latest_block
        self.latest_block = min(db_latest, self.latest_block)
        self.last_processed_block = 0
        logger.debug(f'starting events monitor: latest block number {self.latest_block}')

        self._monitor_is_on = False
        try:
            self._monitor_sleep_time = os.getenv('OCN_SQUID_EVENTS_MONITOR_TIME', 3)
        except ValueError:
            self._monitor_sleep_time = 3

        self._monitor_sleep_time = max(self._monitor_sleep_time, 3)

    @staticmethod
    def get_instance(keeper, storage_path, account):
        if not StarfishEventsManager._instance or StarfishEventsManager._instance.provider_account != account:
            StarfishEventsManager._instance = StarfishEventsManager(keeper, storage_path, account)

        return StarfishEventsManager._instance

    @property
    def db(self):
        return AgreementsStorage(self._storage_path)

    @property
    def provider_account(self):
        return self._account

    @property
    def is_monitor_running(self):
        return self._monitor_is_on

    def start_agreement_events_monitor(self, is_valid_callback=None):
        self._is_valid_callback = is_valid_callback

        if self._monitor_is_on:
            return

        logger.debug(f'Starting the agreement events monitor.')
        t = Thread(
            target=self.run_monitor,
            daemon=True,
        )
        self._monitor_is_on = True
        t.start()
        logger.debug('started the agreement events monitor')

    def stop_monitor(self):
        self._monitor_is_on = False
        self._is_valid_callback = None

    def process_pending_agreements(self, pending_agreements, conditions):
        logger.debug(f'processing pending agreements, there is {len(pending_agreements)} agreements to process.')
        for agreement_id in pending_agreements.keys():
            data = pending_agreements[agreement_id]
            did = data[0]
            consumer_address = data[5]
            block_number = data[6]
            unfulfilled_conditions = conditions[agreement_id].keys()
            self.process_condition_events(
                agreement_id,
                unfulfilled_conditions,
                did,
                consumer_address,
                block_number,
                new_agreement=False
            )

    def get_next_block_range(self):
        to_block = self._web3.eth.blockNumber
        if self.last_processed_block:
            block_range = self.last_processed_block - 1, to_block
        else:
            block_num = self.db.get_latest_block_number() or 0
            if block_num > to_block:
                block_num = to_block - self.last_n_blocks
            from_block = max(to_block - self.last_n_blocks, block_num)
            block_range = from_block, to_block

        logger.debug(f'next block range = {block_range}, latest block number: {to_block}')
        return block_range

    def do_first_check(self):
        db = self.db
        if not db.get_agreement_count():
            return

        block_num = db.get_latest_block_number()
        agreements, conditions = db.get_pending_agreements(block_num - self.last_n_blocks)
        self.process_pending_agreements(agreements, conditions)

    def run_monitor(self):
        try:
            self.do_first_check()
        except Exception as e:
            logger.debug(f'Error processing event: {str(e)}')

        while True:
            try:
                if not self._monitor_is_on:
                    return

                _from, _to = self.get_next_block_range()
                for event_log in self.get_agreement_events(_from, _to):
                    self._handle_agreement_created_event(event_log)

                self.last_processed_block = _to

            except Exception as e:
                logger.debug(f'Error processing event: {str(e)}')

            time.sleep(self._monitor_sleep_time)

    def get_last_block_number(self):
        return self.latest_block - 100

    def get_agreement_events(self, from_block, to_block):
        event_filter = self._keeper.escrow_access_secretstore_template.get_event_filter_for_agreement_created(
            self._account.address, from_block, to_block)
        logger.debug(f'getting event logs in range {from_block} to {to_block} for provider address {self._account.address}')
        logs = event_filter.get_all_entries(max_tries=5)
        # event_filter.uninstall()
        return logs

    def _handle_agreement_created_event(self, event, *_):
        if not event or not event.args:
            return

        if self._account.address != event.args["_accessProvider"]:
            logger.debug(f'agreement event not for my address {self._account.address}, event provider address {event.args["_accessProvider"]}')
            return
        agreement_id = None
        try:
            agreement_id = self._web3.toHex(event.args["_agreementId"])
            ids = self.db.get_agreement_ids()
            if ids:
                # logger.info(f'got agreement ids: #{agreement_id}#, ##{ids}##, \nid in ids: {agreement_id in ids}')
                if agreement_id in ids:
                    logger.debug(f'handle_agreement_created: skipping service agreement {agreement_id} '
                                 f'because it already been processed before.')
                    return

            logger.debug(f'Start handle_agreement_created (agreementId {agreement_id}): event_args={event.args}')

            did = id_to_did(event.args["_did"])

            unfulfilled_conditions = ['lockReward', 'accessSecretStore', 'escrowReward']
            self.process_condition_events(
                agreement_id, unfulfilled_conditions, did, event.args['_accessConsumer'],
                event.blockNumber, new_agreement=True
            )

            logger.debug(f'handle_agreement_created()  (agreementId {agreement_id}) -- '
                         f'done registering event listeners.')

        except Exception as e:
            logger.error(f'Error in handle_agreement_created (agreementId {agreement_id}): {e}', exc_info=1)

    def _last_condition_fulfilled(self, _, agreement_id, cond_name_to_id):
        # update db, escrow reward status to fulfilled
        # log the success of this transaction
        db = self.db
        for cond, _id in cond_name_to_id.items():
            state = self._keeper.condition_manager.get_condition_state(_id)
            db.update_condition_status(agreement_id, cond, state)

        logger.info(f'Agreement {agreement_id} is completed, all conditions are fulfilled.')

    def process_condition_events(self, agreement_id, conditions, did,
                                 consumer_address, block_number, new_agreement=True):

        # check the callback, if set then
        if self._is_valid_callback is not None:
            is_valid = self._is_valid_callback(did, agreement_id, self._account.address, consumer_address)
            if not is_valid:
                return

        ddo = DIDResolver(self._keeper.did_registry).resolve(did)
        service_agreement = ServiceAgreement.from_ddo(ServiceTypes.ASSET_ACCESS, ddo)
        condition_def_dict = service_agreement.condition_by_name
        price = service_agreement.get_price()
        if new_agreement:
            start_time = int(datetime.now().timestamp())
            self.db.record_service_agreement(
                agreement_id, ddo.did, service_agreement.service_definition_id, price,
                ddo.metadata['base'].get('encryptedFiles'), consumer_address, start_time,
                block_number, ServiceTypes.ASSET_ACCESS,
                service_agreement.condition_by_name.keys()
            )

        condition_ids = service_agreement.generate_agreement_condition_ids(
            agreement_id=agreement_id,
            asset_id=ddo.asset_id,
            consumer_address=consumer_address,
            publisher_address=ddo.publisher,
            keeper=self._keeper
        )
        cond_order = ['accessSecretStore', 'lockReward', 'escrowReward']
        cond_to_id = {cond_order[i]: _id for i, _id in enumerate(condition_ids)}
        for cond in conditions:

            if cond == 'lockReward':
                self._keeper.lock_reward_condition.subscribe_condition_fulfilled(
                    agreement_id,
                    max(condition_def_dict['lockReward'].timeout, self.EVENT_WAIT_TIMEOUT),
                    lockRewardCondition.fulfillAccessSecretStoreCondition,
                    (agreement_id, ddo.did, service_agreement, consumer_address,
                     self._account, condition_ids[0]),
                    from_block=block_number
                )
            elif cond == 'accessSecretStore':
                self._keeper.access_secret_store_condition.subscribe_condition_fulfilled(
                    agreement_id,
                    max(condition_def_dict['accessSecretStore'].timeout, self.EVENT_WAIT_TIMEOUT),
                    accessSecretStore.fulfillEscrowRewardCondition,
                    (agreement_id, service_agreement, price, consumer_address, self._account,
                     condition_ids, condition_ids[2]),
                    from_block=block_number
                )
            elif cond == 'escrowReward':
                self._keeper.escrow_reward_condition.subscribe_condition_fulfilled(
                    agreement_id,
                    max(condition_def_dict['escrowReward'].timeout, self.EVENT_WAIT_TIMEOUT),
                    self._last_condition_fulfilled,
                    (agreement_id, cond_to_id),
                    from_block=block_number
                )




"""
   Listing class
"""

from starfish.listing.listing_base import ListingBase
from starfish.account import Account

class Listing(ListingBase):
    """
        Create a Listing object

        :param agent: agent object that created the listing.
        :type agent: :class:`.Agent` object to assign to this Listing
        :param str did: did of the listing, this can be the did of the underling asset.
        :param asset: the core asset for this listing
        :type asset: :class:`.Asset` object
        :param data: data of the listing
        :param ddo: Optional ddo for the listing
        :type data: dict
    """


    def purchase(self, account):
        """

        Purchase the underlying asset within this listing using the account details, return a purchased asset
        with the service_agreement_id ( purchase_id ) set.

        :param account: account to use to purchase this asset.
        :type account: :class:`.Account`

        :return: SquidPurchase object that has information about this listing, asset and purcase id.
        :type: :class:`.SquidPurchase`

        """
        if not isinstance(account, Account):
            raise TypeError('You need to pass an Account object')

        return self._agent.purchase_asset(self, account)

    def is_purchased(self, account):
        """

        Return true if the account has already purchased this listing/asset

        :param account: account to test for this purchase of this asset.
        :type account: :class:`.Account`

        :return: result if has purchased this asset
        :type: boolean

        """
        if not isinstance(account, Account):
            raise TypeError('You need to pass an Account object')

        return self._agent.is_access_granted_for_asset(self.asset, account)

    def set_published(self, value):
        """

        Set the published value

        :params boolean value: Published value True or False

        """
        if self._data and isinstance(self._data, dict) and 'status' in self._data:
            if value:
                self._data['status'] = 'published'
            else:
                self._data['status'] = 'unpublished'

    @property
    def is_published(self):
        """

        Return a True if this listing is published

        :return: True of False if this listing is published
        :type: boolean

        """
        if self._data and isinstance(self._data, dict) and 'status' in self._data:
            return self._data['status'] == 'published'
        return False

    @property
    def get_purchase_ids(self):
        return self._agent.get_asset_purchase_ids(self.asset)

    def __str__(self):
        s = 'Listing: agent=' + self._agent.__class__.__name__ + ', '
        s += 'listing_id=' + self._listing_id + ', '
        s += 'asset=' + self._asset.__class__.__name__ + ', '
        s += 'data=' + str(self._data)
        return s





from starfish.agent.agent_base import AgentBase
from starfish.agent.squid_agent import SquidAgent
from starfish.agent.invoke_agent import InvokeAgent
from starfish.agent.memory_agent import MemoryAgent
from starfish.agent.remote_agent import RemoteAgent



#!/usr/bin/env python
"""

    Tests for `starfish-py account`.

"""
import unittest
import os
import logging
import secrets

from starfish import Ocean, logger
from squid_py import Ocean as SquidOcean
from squid_py import Config as SquidConfig


def test_account_load(ocean, config):
    # address only
    account = ocean.load_account(config.publisher_account['address'], config.publisher_account['password'], config.publisher_account['keyfile'])
    assert(account)
    assert(account.address == config.publisher_account['address'])
    assert(account.password == config.publisher_account['password'])
    assert(account.keyfile == config.publisher_account['keyfile'])

def test_account_get_squid_account(ocean, config):
    account = ocean.load_account(config.publisher_account['address'], config.publisher_account['password'], config.publisher_account['keyfile'])
    assert(account)
    squid_account = account.agent_adapter_account
    assert(squid_account)

def test_account_get_balance(ocean, config):
    account = ocean.load_account(config.publisher_account['address'], config.publisher_account['password'], config.publisher_account['keyfile'])
    assert(account)
    account.request_tokens(1)
    assert(account.ocean_balance)
    assert(account.ether_balance)

"""
def test_account_list(ocean, config):
    publisher_account = ocean.get_account(config.publisher_account)
    accounts = ocean.accounts
    assert(len(accounts) > 1)
    assert(publisher_account.is_address_equal((list(accounts)[0])))
"""






"""
    Operation class
"""

import logging
from starfish.operation.operation_base import OperationBase
from collections import ChainMap
import requests
import json
from starfish.asset import AssetBase

logger = logging.getLogger('starfish.operiation')

class Operation(OperationBase):
    """

    This class returns an operation that can be invoked on a remote Ocean Invoke server
    :param agent: agent that was used create this object.
    :type agent: :class:`.Agent`
    :param did: the did of this Operation
    :type did: str

    """

    def __init__(self, agent,did):
        """init the the Operation Object Base with the agent instance"""
        super().__init__(agent, did)
        self.agent=agent
        self.did=did

    def get_schema(self):
        """
        returns the schema for this particular operation
        """
        r=requests.get(self.agent._koi_url+'operation/'+self.did)
        rawjson=json.loads(r.text)
        self.argnames=[i['name'] for i in rawjson]
        logger.info(f' rawjson {rawjson}')
        self.schema=dict(zip(self.argnames,[i['type'] for i in rawjson]))
        return self.schema

    def _payload(self,account,k,v):
        if self.schema.get(k)=='asset':
            if isinstance(v, AssetBase):
                return {k:{'service_agreement_id':v.purchase_id, 'account':account}}
            raise ValueError("Invalid arguments ")
        else:
            return {k:v}

    def invoke(self, **kwargs):
        """
        Call the invoke function with keyword arguments.
        The keywords should be argument names required in the invoke function payload.
        returns the payload returned by the operation, or ValueError.
        This parameterization of invoke takes only string arguments (and not Ocean assets).
        """
        ## check if all the args are present
        valid_args = len(frozenset(self.argnames).intersection(kwargs.keys()))==len(kwargs)
        ## check if args are not of 'asset' type
        string_args_only = len([k for k,v in kwargs.items() if self.schema.get(k) == 'string']) == len(kwargs)
        logger.info(f' valid args {valid_args} string args {string_args_only}')
        #if True==valid_args and string_args_only==True:
        if True == valid_args:
            r = requests.post(self.agent._koi_url+'freeinvoke/'+self.did,json=kwargs)
            j = json.loads(r.text)
            return j
        else:
            raise ValueError("Invalid arguments ")

    def invoke_op(self,account, **kwargs):
        """
        Call the invoke function with keyword arguments which could have Ocean params (hence invoke_op) The keywords should be argument names required in the invoke function payload.
        returns the payload returned by the operation, or ValueError.
        """
        logger.info(f'calling invoke in operation.py with payload: {kwargs}')
        #test if the keyword arguments are valid
        valid_args=len(frozenset(self.argnames).intersection(kwargs.keys()))==len(kwargs)
        if True == valid_args:
            try:
                payload_seq = [self._payload(account,k,v) for k,v in kwargs.items()]
                payload = dict(ChainMap(*payload_seq))

                r = requests.post(self.agent._koi_url+'freeinvoke/'+self.did,json=payload)
                j = json.loads(r.text)
                return j
            except ValueError:
                raise ValueError(" arguments are invalid according to the schema defn " )
        else:
            raise ValueError("Invalid arguments ")



"""

    Command line console outptu

"""

import json

from .base_output import BaseOutput

class ConsoleOutput(BaseOutput):

    def show(self, text=None, name=None):
        if text:
            print(text)

    def show_item(self, name, value, format_text=None):
        if format_text is None:
            format_text = '{:20}{:80}'
        if isinstance(value, dict):
            value = json.dumps(value)
        print(format_text.format(name, value))

    def show_header(self):
        line = []

        for header in self._header.values():
            line.append(header['format'].format(header['title']))

        if line:
            print("".join(line))

    def show_items(self):
        line = []
        for name, header in self._header.items():
            value = ''
            if name in self._items:
                value = self._items[name]
            line.append(header['format'].format(value))
        if line:
            print("".join(line))

    def flush(self):
        pass



"""
    test_17_asset_purchase

    As a developer working with an Ocean marketplace,
    I want to record a purchase.

"""

import secrets
import logging
import json

from starfish.asset import DataAsset


def test_17_asset_purchase(ocean, resources, config, remote_agent, purchaser_account):
    test_data = secrets.token_hex(1024)
    asset = DataAsset.create('TestAsset', test_data)
    listing = remote_agent.register_asset(asset, resources.listing_data)
    listing.set_published(True)
    logging.debug("create_purchase for listing_id: " + listing.listing_id)
    purchase = remote_agent.purchase_asset(listing, purchaser_account)
    assert(purchase['listingid'] == listing.listing_id)
    assert(purchase['status'] == 'wishlist')
    logging.debug("purchase: " + json.dumps(purchase))



from unittest.mock import Mock
import pytest
import secrets

from tests.unit.libs.unit_test_config import unitTestConfig
from starfish import Ocean
from tests.unit.mocks.mock_squid_agent_adapter import MockSquidAgentAdapter



@pytest.fixture(scope="module")
def ocean():
    result = Ocean(keeper_url=unitTestConfig.keeper_url,
            contracts_path=unitTestConfig.contracts_path,
            gas_limit=unitTestConfig.gas_limit,
            squid_agent_adapter_class=MockSquidAgentAdapter,
            connect=False
    )
    return result

@pytest.fixture(scope="module")
def config():
    return unitTestConfig



"""

Test asset modules in __init__.py

"""
import secrets
import logging
import json

from starfish.asset import (
    create_asset_from_metadata_text,
    AssetBase,
    BundleAsset,
    DataAsset,
    OperationAsset,
)


def test_create_asset_from_metadata():

    type_list = [
        ( 'bundle', BundleAsset ),
        ( 'dataset', DataAsset ),
        ( 'operation', OperationAsset ),
    ]
    for type_name, class_type in type_list:
        metadata = {
            'type': type_name,
            'name': f'TestAsset_{type_name}'
        }
        asset = create_asset_from_metadata_text(json.dumps(metadata))
        assert(asset)
        assert(isinstance(asset, class_type))



"""

Agent class to provide basic functionality for all Ocean Agents

"""


import datetime
import json
import logging
import re
import traceback
import sys

from web3 import Web3

from starfish.middleware.squid_agent_adapter import (
    SquidAgentAdapter,
    SquidAgentAdapterPurchaseError
)
from starfish.account import Account
from starfish.agent import AgentBase
from starfish.listing import Listing
from starfish.asset import (
    BundleAsset,
    DataAsset,
    RemoteDataAsset,
 )
from starfish.purchase import Purchase
from starfish.utils.did import did_parse
from starfish.exceptions import (
    StarfishAssetNotFound,
    StarfishPurchaseError
)

from squid_py.brizo.brizo_provider import BrizoProvider
from ocean_utils.ddo.metadata import (
    MetadataBase,
    AdditionalInfoMeta,
)
from ocean_keeper.exceptions import OceanDIDNotFound


logger = logging.getLogger('starfish.squid_agent')


ALLOWED_FILE_META_ITEMS = [
    'index',
    'url',
    'encoding',
    'compression',
    'checksum',
    'checksumType',
    'contentLength',
    'contentType',
    'resourceId',
]

class SquidAgent(AgentBase):
    """

    Squid Agent class allows to register and list asset listings.

    :param ocean: Ocean object that is being used.
    :type ocean: :class:`.Ocean`
    :param aquarius_url: Aquarius url ( http://localhost:5000 ).
    :type aquarius_url: str or None
    :param brizo_url: Brizo url (http://localhost:8030).
    :type brizo_url: str or None
    :param secret_store_url: Secret store URL (http://localhost:8010).
    :type secret_store_url: str or None
    :param parity_url: Parity URL, if you are using the secret store (http://localhost:9545').
    :type parity_url: str or None
    :param storage_path: Path to the storage db (squid_py.db).
    :type storage_path: str or None

    Example how to use this agent: ::

        # First import the classes
        from starfish.agent import SquidAgent
        from starfish import Ocean
        from starfish.asset import RemoteDataAsset

        # create the ocean object
        ocean = Ocean()

        # get your publisher account
        account = ocean.load_account('0x00bd138abd70e2f00903268f3db08f2d25677c9e')

        #create the SquidAgent
        my_config = {
            'aquarius_url': 'http://localhost:5000',
            'brizo_url': 'http://localhost:8030',
            'secret_store_url': 'http://localhost:12001',
            'parity_url': 'http://localhost:9545',
            'storage_path': 'squid_py.db',
        }
        agent = SquidAgent(ocean, my_config)

        asset = RemoteDataAsset.create_with_url('MyAsset', 'http://my_asset_url.com')
        # register an asset data and listing info
        listing = agent.register(asset, account)
    """

    def __init__(self, ocean, *args, **kwargs):
        """init a standard ocean object"""
        AgentBase.__init__(self, ocean)
        self._adapter = None

        if args and isinstance(args[0], dict):
            kwargs = args[0]

        self._aquarius_url = kwargs.get('aquarius_url', 'http://localhost:5000')
        self._brizo_url = kwargs.get('brizo_url', 'http://localhost:8030')
        self._secret_store_url = kwargs.get('secret_store_url', 'http://localhost:12001')
        self._storage_path = kwargs.get('storage_path', 'squid_py.db')
        self._parity_url = kwargs.get('parity_url', self._ocean.keeper_url)

    def register_asset(self, asset, listing_data, account):
        """

        Register a squid asset with the ocean network.

        :param asset: the asset to register, at the moment only a Asset can be used.
        :type asset: :class:`.DataAsset`, :class:`.RemoteAsset` or :class:`.BundleAsset` object to register
        :param dict listing_data: data that is required for listing a registered asset
        :param account: Ocean account to use to register this asset.
        :type account: :class:`.Account` object to use for registration.
        :param dict options:

        :return: A new :class:`.Listing` object that has been registered, if failure then return None.
        :type: :class:`.Listing` class

        At the moment only support :class:`.DataAsset` or :class:`.BundleAsset`

        For example::

            metadata = json.loads('my_metadata')
            # get your publisher account
            account = ocean.get_account('0x00bd138abd70e2f00903268f3db08f2d25677c9e')
            agent = SquidAgent(ocean)
            asset = DataAsset.create_from_file('My test asset', 'Testfile.txt')
            listing = agent.register_asset(asset, {'price': 8}, account)

            if listing:
                print(f'registered my listing asset for sale with the did {listing.did}')

        """

        if not isinstance(account, Account):
            raise TypeError('You need to pass an Account object')

        if not account.is_valid:
            raise ValueError('You must set the account address, password and keyfile')

        if not isinstance(listing_data, dict):
            raise TypeError('You must provide some listing data as dict')

        if not SquidAgent.is_supported_asset(asset):
            raise TypeError('This agent only supports a DataAsset, RemoteDataAsset or BundleAsset')

        adapter = self.agent_adapter
        account.agent_adapter = adapter
        metadata = SquidAgent._convert_listing_asset_to_metadata(asset, listing_data)

        ddo = adapter.register_asset(metadata, account.agent_adapter_account)

        listing = None
        if ddo:
            asset.set_did(ddo.did)
            listing = Listing(self, ddo.did, asset, listing_data, ddo)

        return listing


    def validate_asset(self, asset):
        """

        Validate an asset

        :param asset: Asset to validate.
        :return: True if the asset is valid
        """

        if not asset:
            raise ValueError('asset must be an object')
        if not SquidAgent.is_supported_asset(asset):
            raise TypeError('asset must be a type of DataAsset, RemoteDataAsset or BundleAsset object')
        if not asset.metadata:
            raise ValueError('Metadata must have a value')
        if not isinstance(asset.metadata, dict):
            raise ValueError('Metadat must be a dict')

#        adapter = self.agent_adapter
#        return adapter.validate_metadata(asset.metadata)
        return True

    def get_listing(self, listing_id):
        """
        Return an listing from the given listing_id.

        :param str listing_id: Id of the listing.

        :return: a registered listing given a Id of the listing
        :type: :class:`.Listing` class

        :raise: :class:`.StarfishAssetNotFound` if the asset is not found in
            aquarius or the DID of the asset is not on the network ( Block chain )
        """
        listing = None
        adapter = self.agent_adapter

        try:
            ddo = adapter.read_asset(listing_id)
        except OceanDIDNotFound as e:
            raise StarfishAssetNotFound(e)

        if ddo:
            listing = self._listing_from_ddo(ddo)
        return listing


    def search_listings(self, text, sort=None, offset=100, page=1):
        """

        Search the off chain storage for an asset with the givien 'text'

        :param str, dict text: Test to search all metadata items for. If the field is a string
            then the search will be a basic search, else a dict then a query will be performed
        :param sort: sort the results ( defaults: None, no sort).
        :type sort: str or None
        :param int offset: Return the result from with the maximum record count ( defaults: 100 ).
        :param int page: Returns the page number based on the offset ( defaults: 1 ).

        :return: a list of assets objects found using the search.
        :type: list of DID strings

        For example::

            # return the 300 -> 399 records in the search for the text 'weather' in the metadata.
            my_result = agent.search_registered_assets('weather', None, 100, 3)

        """
        adapter = self.agent_adapter
        if isinstance(text, str) or isinstance(text, dict):
            ddo_list = adapter.search_assets(text, sort, offset, page)
        else:
            raise ValueError('You can only pass a str or dict for the search text')
        result = []
        for ddo in ddo_list:
            listing = self._listing_from_ddo(ddo)
            result.append(listing)

        return result

    def purchase_asset(self, listing, account):
        """

        Purchase an asset using it's listing and an account.

        :param listing: Listing to use for the purchase.
        :type listing: :class:`.Listing`
        :param account: Ocean account to purchase the asset.
        :type account: :class:`.Account` object to use for registration.

        :raise: :class:`.StarfishAssetNotFound` if the asset is not found in
            aquarius or the DID of the asset is not on the network ( Block chain )

        """
        purchase = None
        adapter = self.agent_adapter
        account.agent_adapter = adapter

        try:
            if 'price' in listing.data:
                account_balance = account.ocean_balance
                asset_price = listing.data['price']
                if account_balance < asset_price:
                    raise StarfishPurchaseError(f'Insufficient Funds: Your account balance has {account_balance} which is not enougth to purchase the asset at a price of {asset_price}')

            service_agreement_id = adapter.purchase_asset(listing.ddo, account.agent_adapter_account)
        except OceanDIDNotFound as e:
            raise StarfishAssetNotFound(e)

        purchase = Purchase(self, listing, service_agreement_id, account)

        return purchase

    def is_access_granted_for_asset(self, asset, account, purchase_id=None):
        """

        Check to see if the account and purchase_id have access to the assed data.


        :param asset: Asset to check for access.
        :type asset: :class:`.Asset` object
        :param account: Ocean account to purchase the asset.
        :type account: :class:`.Account` object to use for registration.
        :param str purchase_id: Optional purchase id that was used to purchase the asset.

        :return: True if the asset can be accessed and consumed.
        :type: boolean
        """

        adapter = self.agent_adapter
        account.agent_adapter = adapter
        if purchase_id:
            return adapter.is_access_granted_for_asset(asset.did, account.agent_adapter_account, purchase_id)
        else:
            purchase_id_list = adapter.get_asset_purchase_ids(asset.did)
            for purchase_id in purchase_id_list:
                if adapter.is_access_granted_for_asset(asset.did, account.agent_adapter_account, purchase_id):
                    return True
        return False

    def get_asset_purchase_ids(self, asset):
        """

        Returns as list of purchase id's that have been used for this asset

        :param asset: Asset to return purchase details.
        :type asset: :class:`.DataAsset` or :class:`.BundleAsset` object

        :return: list of purchase ids
        :type: list

        """
        adapter = self.agent_adapter

        return adapter.get_asset_purchase_ids(asset.did)



    def purchase_wait_for_completion(self, asset, account, purchase_id, timeoutSeconds):
        """

        Wait for completion of the purchase

        TODO: issues here...
        + No method as yet to pass back paramaters and values during the purchase process
        + We assume that the following templates below will always be used.


        :param integer timeoutSeconds: Optional seconds to waif to purchase to complete. Default: 60 seconds
        :return: True if successfull or an error message if failed
        :type: string or boolean

        :raises OceanPurchaseError: if the correct events are not received

        """
        adapter = self.agent_adapter
        account.agent_adapter = adapter
        if not purchase_id:
            raise ValueError('Please provide a valid purhase id')

        try:
            adapter.purchase_wait_for_completion(asset.did, account.agent_adapter_account, purchase_id, timeoutSeconds)
        except SquidAgentAdapterPurchaseError as purchaseError:
            raise StarfishPurchaseError(purchaseError)
        except Exception as e:
            raise e
        return True


    def consume_asset(self, listing, account, purchase_id):
        """

        Consume the asset and download the data. The actual payment to the asset
        provider will be made at this point.

        :param listing: Listing that was used to make the purchase.
        :type listing: :class:`.Listing`
        :param str purchase_id: purchase id that was used to purchase the asset.
        :param account: Ocean account that was used to purchase the asset.
        :type account: :class:`.Account` object to use for registration.
        :param str download_path: path to store the asset data.

        :return: BunleAsset object if you have purchased this asset, which contains one or more RemoteAssets
        :type: :class:`.BundleAsset`

        """
        asset = None
        adapter = self.agent_adapter
        account.agent_adapter = adapter
        file_list = adapter.consume_asset(listing.ddo, account.agent_adapter_account, purchase_id)
        if file_list:
            asset = BundleAsset.create('SquidAssetBundle', did=listing.ddo.did)
            for index, file_item in enumerate(file_list):
                asset_item = SquidAgent.create_data_asset_from_file_item(index, file_item, listing.ddo.did)
                asset.add(asset_item.name, asset_item)
        return asset

    def start_agreement_events_monitor(self, account, callback=None):
        """

        The provider or publisher needs to watch the events on the block chain, to
        then setup the agreement contracts on the block chain when the consume
        first starts to make the initial consume request ( payment to escrow ).

        :param account: Account to watch for payment events for the provider of the asset
        :type account: :class:`.Account`

        """
        adapter = self.agent_adapter
        account.agent_adapter = adapter
        adapter.start_agreement_events_monitor(account.agent_adapter_account, callback)

    def stop_agreement_events_monitor(self):
        adapter = self.agent_adapter
        adapter.stop_agreement_events_monitor()

    def _listing_from_ddo(self, ddo):
        """ convert a ddo to a listing that contains a BundleAsset """

        listing_data, asset_metadata = self._convert_ddo_to_listing_data_asset_metadata(ddo)
        asset = BundleAsset.create('SquidAssetBundle', did=ddo.did)
        for index, file_item in enumerate(asset_metadata):
            asset_item = SquidAgent.create_data_asset_from_file_item(index, file_item, ddo.did)
            asset.add(asset_item.name, asset_item)
        listing_id = ddo.did
        listing = Listing(self, listing_id, asset, listing_data, ddo)
        return listing

    @staticmethod
    def _convert_ddo_to_listing_data_asset_metadata(ddo):
        listing_data = ddo.metadata.get(MetadataBase.KEY, None)
        # put back additional fields that cannot be saved with the squid
        # main metadata

        # first copy the price from the network as the '' price ( in vodka's )
        listing_data['price_vodka'] = int(listing_data['price'])
        # all starfish prices are in ocean tokens - so convert from Vodka's to Ocean tokens
        listing_data['price'] = Web3.fromWei(int(ddo.metadata[MetadataBase.KEY]['price_vodka']), 'ether')
        asset_metadata = []
        if listing_data['files']:
            for file_data in listing_data['files']:
                asset_metadata.append(file_data)

        info = ddo.metadata.get(AdditionalInfoMeta.KEY, None)
        if info and isinstance(info, dict):
            for name, value in info.items():
                # for convinience we will try to decode 'extra_data' field as a json to dict
                try:
                    if name == 'extra_data' and isinstance(value, str):
                        value = json.loads(value)
                except json.decoder.JSONDecodeError:
                    pass
                listing_data[name] = value

        return listing_data, asset_metadata

    @staticmethod
    def _convert_listing_asset_to_metadata(asset, listing_data):
        """
        For squid we need to create a single metadata record from a listing_data and asset/s
        """
        metadata = {
            MetadataBase.KEY: {
                'name': 'SquidAsset',
                'type': 'dataset',
                'dateCreated': datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ'),
                'author': 'Author',
                'license': 'closed',
                'price': '0',
                'files': []
            },
            AdditionalInfoMeta.KEY: {}
        }
        if 'extra_data' in listing_data:
            metadata[AdditionalInfoMeta.KEY]['extra_data'] = json.dumps(listing_data['extra_data'])
            del listing_data['extra_data']

        for name, value in listing_data.items():
            if name in MetadataBase.VALUES_KEYS:
                metadata[MetadataBase.KEY][name] = value
            else:
                if name != 'extra_data':
                    metadata[AdditionalInfoMeta.KEY][name] = value

        # make sure we are sending a price value as string for squid in vodka
        price_value = Web3.toWei(metadata[MetadataBase.KEY]['price'], 'ether')
        metadata[MetadataBase.KEY]['price'] = str(price_value)

        def append_asset_file(metadata, asset):
            asset_metadata = asset.metadata
            metadata['base']['files'].append(asset_metadata)
            index = len(metadata['base']['files']) - 1
            metadata['base']['files'][index]['index'] = index
            url = None
            if 'url' in asset_metadata:
                url = asset_metadata['url']
            elif 'filename' in asset_metadata:
                url = 'file://' + asset_metadata['filename']
                del asset_metadata['filename']
            if url:
                metadata['base']['files'][index]['url'] = url
            else:
                raise ValueError(f'The DataAsset does not contain a "url" or "filename" metadata item')

        if isinstance(asset, BundleAsset):
            for _, asset_item in asset:
                if not (isinstance(asset_item, DataAsset)):
                    raise TypeError(f'Invalid asset type {type(asset_item)}: The BundleAsset can only contain multilple assets of the type DataAsset')
                append_asset_file(metadata, asset_item)
        else:
            append_asset_file(metadata, asset)

        for item in metadata['base']['files']:
            delete_list = []
            for name in item.keys():
                if name not in ALLOWED_FILE_META_ITEMS:
                    delete_list.append(name)
            for name in delete_list:
                metadata[AdditionalInfoMeta.KEY][f'file_{name}'] = item[name]
                del item[name]
        return metadata


    @staticmethod
    def invoke_operation(listing, account, purchase_id, payload):
        """
        Invoke the operation

        :param listing: Listing that was used to make the purchase.
        :type listing: :class:`.Listing`
        :param account: Ocean account that was used to purchase the asset.
        :type account: :class:`.Account` object to use for registration.
        :param str purchase_id: purchase id that was used to purchase the asset.
        :param str payload: params required for the operation

        :return: True if the operation was invoked
        :type: boolean

        """
        logger.info(f'calling invoke in squid_agent.py with payload: {payload}')
        try :
            ddo= listing.data
            files = ddo.metadata['base']['encryptedFiles']
            logger.info(f'encrypted contentUrls: {files}')
            files = files if isinstance(files, str) else files[0]
            sa = SquidAgentAdapter.get_service_agreement_from_ddo(ddo)
            service_url = sa.service_endpoint
            if not service_url:
                logger.error(
                    'Consume asset failed, service definition is missing the "serviceEndpoint".')
                raise AssertionError(
                    'Consume asset failed, service definition is missing the "serviceEndpoint".')

            return BrizoProvider.get_brizo().invoke_service(
                purchase_id, service_url,
                account.address, payload)
        except Exception:
            logger.error('got error invoking Brizo')
            traceback.print_exc(file=sys.stdout)
            return False



    @property
    def agent_adapter(self):
        """

        Return an instance of the squid ageent adapter, for access to the squid library layer
        :return: squid Agent Adapter object
        """

        if not self._adapter:
            options = {
                'aquarius_url': self._aquarius_url,
                'brizo_url': self._brizo_url,
                'secret_store_url': self._secret_store_url,
                'parity_url': self._parity_url,
                'storage_path': self._storage_path,
            }
            self._adapter = self._ocean.get_squid_agent_adapter(options)
        return self._adapter

    @staticmethod
    def is_did_valid(did):
        """
        Checks to see if the DID string is a valid DID for this type of Asset.
        This method only checks the syntax of the DID, it does not resolve the DID
        to see if it is assigned to a valid Asset.

        :param str did: DID string to check to see if it is in a valid format.

        :return: True if the DID is in the format 'did:op:xxxxx'
        :type: boolean
        """
        data = did_parse(did)
        return not data['path']

    @staticmethod
    def create_data_asset_from_file_item(index, file_item, did):
        metadata = {
            'type': 'dataset',
            'name': f'SquidAsset_{index}',
        }

        for name in ALLOWED_FILE_META_ITEMS:
            if name in file_item:
                metadata[name] = file_item[name]
                if name == 'url':
                    url = file_item['url']
                    if re.match('^file://', url):
                        metadata['filename'] = re.sub('^file://', '', url)

        return RemoteDataAsset(metadata, did=did)

    @staticmethod
    def is_supported_asset(asset):
        """

        Return True if the asset is a supported type.

        :param obj asset: Asset object to check to see if it is the correct type

        :return: True if the type is supported

        """
        return (isinstance(asset, DataAsset) or isinstance(asset, RemoteDataAsset) or isinstance(asset, BundleAsset))



"""
    test_pd_case_file_transfer

    As a developer
    I want save an asset file on remote agent, and sale the asset via squid,
    As a consumer
    I want to purchase an asset vi Ocean and download the asset from Sufer.

"""

import secrets
import logging
import json
import base64

from starfish.asset import (
    BundleAsset,
    DataAsset,
    RemoteDataAsset,
)

def test_pd_case_file_transfer(ocean, config, resources, remote_agent,
    squid_agent, publisher_account, purchaser_account):

    # take copy of the stored data to compare later
    with open(resources.asset_file, 'rb') as fp:
        store_data = fp.read()

    # save the asset file to remote agent
    asset_store = DataAsset.create_from_file('TestAsset', resources.asset_file)
    listing_store = remote_agent.register_asset(asset_store, resources.listing_data)
    assert(listing_store)

    # now upload to the storage
    remote_agent.upload_asset(asset_store)

    # now register the asset link to remote agent in squid
    download_link = asset_store.did
    resourceId = base64.b64encode(bytes(resources.asset_file)).decode('utf-8')
    asset_sale = RemoteDataAsset.create_with_url('SquidAsset', download_link, metadata={'resourceId': resourceId})
    # print('metadata ',squid_agent._convert_listing_asset_to_metadata(asset_sale, resources.listing_data))
    listing = squid_agent.register_asset(asset_sale, resources.listing_data, account=publisher_account)
    assert(listing)

    # now re-read the listing to make sure that we get the same result and listing data
    listing = squid_agent.get_listing(listing.listing_id)

    # now start the purchase part
    # setup the purchase account
    logging.info(f'purchaser_account {purchaser_account.ocean_balance}')
    # request the tokens to buy the asset
    purchaser_account.request_tokens(10)

    squid_agent.start_agreement_events_monitor(publisher_account)


    # purchase the linked remote asset
    purchase = listing.purchase(purchaser_account)
    assert(purchase)

    assert(not purchase.is_completed)

    # wait for completion of purchase
    error_message = purchase.wait_for_completion()
    assert(error_message == True)

    # check to see if purchased
    assert(purchase.is_completed)


    assert(purchase.is_purchased)
    assert(purchase.is_purchase_valid)

    # get the purchased asset from squid
    purchase_asset = purchase.consume_asset
    assert(purchase_asset)
    # this is a bundle asset with a collection of remote assets
    assert(isinstance(purchase_asset, BundleAsset))

    # we are only using the first asset, so get it from the bundle
    remote_asset = purchase_asset.get_asset_at_index(0)
    assert(isinstance(remote_asset, RemoteDataAsset))

    #get the remote agent did and asset_id from the 'url'
    assert(remote_asset.url)

    remote_agent_did, asset_id = remote_agent.decode_asset_did(remote_asset.url)
    assert(remote_agent_did)
    assert(asset_id)

    # get the actual URL of the remote agent, and asset storage component
    download_url = remote_agent.get_asset_store_url(asset_id)
    assert(download_url)

    # download the asset from storage
    new_asset_store = remote_agent.download_asset(asset_id, download_url)
    assert(new_asset_store)
    assert(new_asset_store.is_asset_type('dataset'))

    # final check stored asset data is == to original data put up for sale
    assert(new_asset_store.data == store_data)

    # check the resource id in the purchased asset
    assert('resourceId' in remote_asset.metadata)
    asset_file_path = base64.b64decode(remote_asset.metadata['resourceId']).decode('utf-8')
    assert(str(resources.asset_file) == asset_file_path)




"""
   Listing Base class
"""

from abc import ABC, abstractmethod

class ListingBase(ABC):
    """
        Create a Listing object

        :param agent: agent object that created the listing.
        :type agent: :class:`.Agent` object to assign to this Listing
        :param str listing_id: id of the listing.
        :param asset: the core asset for this listing
        :type asset: :class:`.Asset` object
        :param data: data of the listing
        :param ddo: Optional DDO object for this listing
        :type data: dict
    """
    def __init__(self, agent, listing_id, asset, data, ddo=None):
        """init the the Listing Object Base with the agent instance"""
        self._agent = agent
        self._listing_id=listing_id
        self._asset = asset
        self._data = data
        self._ddo = ddo
        super().__init__()

    @abstractmethod
    def purchase(self, account):
        """

        Purchase the underlying asset within this listing using the account details, return a purchased asset
        with the service_agreement_id ( purchase_id ) set.

        :param account: account to use to purchase this asset.
        :type account: :class:`.Account`

        :return: SquidPurchase object that has information about this listing, asset and purcase id.
        :type: :class:`.SquidPurchase`

        """
        pass

    @property
    def agent(self):
        """

        :return: Agent object that created this listing
        :type: :class:`.SquidAgent`
        """
        return self._agent

    @property
    def data(self):
        """

        :return: data of the listing
        :type: dict or None
        """
        return self._data

    @property
    def ddo(self):
        """

        :return: ddo of the listing
        :type: dict or None
        """
        return self._ddo

    @property
    def listing_id(self):
        """

        :return: the listing id
        :type: str or None
        """
        return self._listing_id

    @property
    def asset(self):
        """

        :return: asset held by the listing
        :type: :class:`.Asset`
        """
        return self._asset

    @property
    def did(self):
        """

        :return: asset held did
        :type: str
        """
        if self._asset:
            return self._asset.did

    @property
    def is_empty(self):
        """

        Checks to see if this Listinng is empty.

        :return: True if this listing is empty else False.
        :type: boolean
        """
        return self._listing_id is None or self._asset is None



"""
    test_16_asset_bundle

    As a developer working with asset bundles
    I want a way to get to sub-assets.

"""

import secrets
import logging
import json

from starfish.asset import (
    DataAsset,
    BundleAsset,
)

TEST_ASSET_COUNT = 4


def test_16_asset_bundle_register(ocean, config, remote_agent, metadata):
    pass




import os

from configparser import (
    ConfigParser,
    ExtendedInterpolation,
)


class IntegrationTestConfig():
    def __init__(self, filename):

        config = ConfigParser(interpolation=ExtendedInterpolation())
        config.read(filename)

        if 'BARGE_URL' in os.environ:
            config.set('test', 'barge_url', os.environ['BARGE_URL'])

        self.keeper_url = config.get('ocean', 'keeper_url')
        self.contracts_path = config.get('ocean', 'contracts_path')
        self.gas_limit = config.get('ocean', 'gas_limit')

        self.publisher_account = {
            'address': config.get('publisher account', 'address'),
            'password': config.get('publisher account', 'password'),
            'keyfile': config.get('publisher account', 'keyfile'),
        }
        self.purchaser_account = {
            'address': config.get('purchaser account', 'address'),
            'password': config.get('purchaser account', 'password'),
            'keyfile': config.get('purchaser account', 'keyfile'),
        }
        self.agent_account = {
            'address': config.get('agent account', 'address'),
            'password': config.get('agent account', 'password'),
            'keyfile': config.get('agent account', 'keyfile'),
        }

        items = config.items('squid agent')
        self.squid_config = {}
        for item in items:
            self.squid_config[item[0]] = item[1]

        self.remote_agent_username=config.get('remote agent', 'username')
        self.remote_agent_password=config.get('remote agent', 'password')
        self.remote_agent_url=config.get('remote agent', 'url')

        self.koi_url=config.get('invoke agent', 'url')




import datetime
import pytest
import secrets
import tempfile
import math


from starfish.agent.squid_agent import SquidAgent
from starfish.asset import (
    BundleAsset,
    DataAsset,
    RemoteDataAsset,
)
from starfish.exceptions import StarfishPurchaseError
from tests.unit.mocks.mock_squid_agent_adapter import MockSquidAgentAdapter


VALID_DID = 'did:op:' + secrets.token_hex(64)
INVALID_DID = 'did:ox:' + secrets.token_hex(128)
TEST_INIT_PARMS = {
    'aquarius_url': 'http://test_aquarius:5000',
    'brizo_url': 'http://test_brizo:8030',
    'secret_store_url': 'http://test_secret_store_url:12001',
    'storage_path': 'test_squid_path',
}

def _register_asset(ocean, resources, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    agent = SquidAgent(ocean)
    assert(agent)
    asset = RemoteDataAsset.create_with_url('test url asset', resources.asset_remote)
    assert(asset)
    listing = agent.register_asset(asset, resources.listing_data, account)
    return (listing, agent, asset)

def _purchase_asset(ocean, resources, config):
    listing, agent, asset = _register_asset(ocean, resources, config)
    account = ocean.load_account(config.accounts[1].as_dict)
    purchase = agent.purchase_asset(listing, account)
    return purchase, listing, agent, asset, account

def test_init(ocean):

    agent = SquidAgent(ocean, TEST_INIT_PARMS)
    assert(agent)
    assert isinstance(agent.agent_adapter, MockSquidAgentAdapter)
    adapter = agent.agent_adapter
    assert(adapter.options['aquarius_url'] == TEST_INIT_PARMS['aquarius_url'])
    assert(adapter.options['brizo_url'] == TEST_INIT_PARMS['brizo_url'])
    assert(adapter.options['secret_store_url'] == TEST_INIT_PARMS['secret_store_url'])
    assert(adapter.options['storage_path'] == TEST_INIT_PARMS['storage_path'])

    agent = SquidAgent(ocean,
        aquarius_url = TEST_INIT_PARMS['aquarius_url'],
        brizo_url = TEST_INIT_PARMS['brizo_url'],
        secret_store_url = TEST_INIT_PARMS['secret_store_url'],
        storage_path = TEST_INIT_PARMS['storage_path']
    )
    assert(agent)
    adapter = agent.agent_adapter
    assert(adapter)
    assert(adapter.options['aquarius_url'] == TEST_INIT_PARMS['aquarius_url'])
    assert(adapter.options['brizo_url'] == TEST_INIT_PARMS['brizo_url'])
    assert(adapter.options['secret_store_url'] == TEST_INIT_PARMS['secret_store_url'])
    assert(adapter.options['storage_path'] == TEST_INIT_PARMS['storage_path'])

def test_register_asset(ocean, resources, config):
    listing, agent, asset = _register_asset(ocean, resources, config)
    assert(listing)
    assert(listing.listing_id)

def test_register_bundle_asset(ocean, resources, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    agent = SquidAgent(ocean)
    assert(agent)
    bundle_asset = BundleAsset.create('test bundle asset')
    for index in range(0, 5):
        asset = RemoteDataAsset.create_with_url('test remote asset', resources.asset_remote)
        assert(asset)
        bundle_asset.add(f'name_{index}', asset)
    listing = agent.register_asset(bundle_asset, resources.listing_data, account)
    assert(listing)

def test_register_asset_invalid_data(ocean, resources, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    agent = SquidAgent(ocean)
    assert(agent)
    asset = DataAsset.create('test remote asset', 'some data that is never going to be saved')
    with pytest.raises(ValueError):
        listing = agent.register_asset(asset, resources.listing_data, account)

def test_validate_asset(ocean):
    agent = SquidAgent(ocean, TEST_INIT_PARMS)
    assert(agent)
    asset = RemoteDataAsset.create_with_url('TestAsset', 'http://dex.sg')
    assert(agent.validate_asset(asset))

def test_get_listing(ocean, resources, config):
    listing, agent, asset = _register_asset(ocean, resources, config)
    found_listing = agent.get_listing(listing.listing_id)
    assert(found_listing)
    assert(found_listing.listing_id == listing.listing_id)

def test_search_listings(ocean, resources, config):
    listing, agent, asset = _register_asset(ocean, resources, config)
    listing_items = agent.search_listings(resources.listing_data['author'])
    assert(listing_items)
    assert(len(listing_items) > 0)
    is_found = False
    listing_id = listing.listing_id
    for listing in listing_items:
        if listing_id == listing.listing_id:
            is_found = True
            break
    assert(is_found)

def test_purchase_asset(ocean, resources, config):
    listing, agent, asset = _register_asset(ocean, resources, config)
    account = ocean.load_account(config.accounts[1].as_dict)
    purchase = agent.purchase_asset(listing, account)
    assert(purchase)

def test_is_access_granted_for_asset(ocean, resources, config):
    purchase, listing, agent, asset, account = _purchase_asset(ocean, resources, config)
    assert(agent.is_access_granted_for_asset(asset, account, purchase.purchase_id, ))

def test_purchase_wait_for_completion(ocean, resources, config):
    purchase, listing, agent, asset, account = _purchase_asset(ocean, resources, config)
    assert(agent.purchase_wait_for_completion(asset, account, purchase.purchase_id, 30))
    # test raised error if purchase failed
    with pytest.raises(ValueError):
        agent.purchase_wait_for_completion(asset, account, None, 30)

def test_consume_asset(ocean, resources, config):
    purchase, listing, agent, asset, account = _purchase_asset(ocean, resources, config)
    assert(agent.consume_asset(listing, account,  purchase.purchase_id))

def test_is_did_valid():
    assert(SquidAgent.is_did_valid(VALID_DID))
    # FIXME: This should be invalid
    assert( SquidAgent.is_did_valid(INVALID_DID))

def test_price_out_of_range(ocean, resources, config):
    account = ocean.load_account(config.accounts[0].as_dict)
    agent = SquidAgent(ocean)
    assert(agent)
    asset = RemoteDataAsset.create_with_url('test squid asset with url', resources.asset_remote)
    assert(asset)
    resources.listing_data['price'] = -1
    with pytest.raises(ValueError):
        listing = agent.register_asset(asset, resources.listing_data, account)
    resources.listing_data['price'] = math.pow(2, 197)
    with pytest.raises(ValueError):
        listing = agent.register_asset(asset, resources.listing_data, account)



"""


    test_01_connect_to_ocean


    As a developer working with Ocean,
    I need a way to connect to the Ocean Network

"""

import secrets
import json
from starfish import Ocean
from starfish.agent import RemoteAgent
from starfish.utils.did import did_generate_random


def test_01_connect_to_ocean(ocean, config):
    assert(ocean)


def test_01_connect_to_remote_agent(ocean, config, publisher_account):

    did = did_generate_random()
    assert(did)
    ddo = {
        'name': 'Test ddo',
        'value': secrets.token_hex(64)
    }

    ocean.register_did(did, json.dumps(ddo), publisher_account)
    resolve_ddo = ocean.resolve_did(did)
    assert(resolve_ddo)
    found_ddo = json.loads(resolve_ddo)
    assert(found_ddo == ddo)

def test_01_connect_to_multiple_remote_agent(ocean, config, publisher_account):

    for index in range(0, 2):
        did = did_generate_random()
        assert(did)
        ddo = {
            'name': 'Test ddo',
            'value': secrets.token_hex(64),
            'index': index
        }
        ocean.register_did(did, json.dumps(ddo), publisher_account)
        resolve_ddo = ocean.resolve_did(did)
        assert(resolve_ddo)
        found_ddo = json.loads(resolve_ddo)
        assert(found_ddo == ddo)



"""
DID utils

"""

import secrets
import re
from urllib.parse import urlparse
from web3 import Web3

OCEAN_DID_METHOD = 'dep'

def did_parse(did):
    """parse a DID into it's parts"""
    if not isinstance(did, str):
        raise TypeError('Expecting DID of string type, got %s of %s type' % (did, type(did)))

    match = re.match('^did:([a-z0-9]+):([a-zA-Z0-9-.]+)(.*)', did)
    if not match:
        raise ValueError('DID %s does not seem to be valid.' % did)

    result = {
        'method': match.group(1),
        'id': match.group(2),
        'path': None,
        'fragment': None,
        'id_hex': None
    }

    uri_text = match.group(3)
    if uri_text:
        uri = urlparse(uri_text)
        result['fragment'] = uri.fragment
        if uri.path:
            result['path'] = uri.path[1:]

    if result['method'] == OCEAN_DID_METHOD and re.match('^[0-9A-Fa-f]{1,64}$', result['id']):
        result['id_hex'] = Web3.toHex(hexstr=result['id'])

    if not result['id_hex'] and result['id'].startswith('0x'):
        result['id_hex'] = result['id']

    return result

def did_generate_random():
    did_id = secrets.token_hex(32)
    return f'did:{OCEAN_DID_METHOD}:{did_id}'


def did_to_asset_id(did):
    data = did_parse(did)
    asset_id = None
    if data['id_hex']:
        asset_id = data['id_hex']
    if data['path'] and re.match('^[0-9A-Fa-f]{1,64}$', data['path']):
        asset_id = '0x' + data['path']
    return asset_id



"""

    Tests for `squid-py issues` for seaching tags in metadata.

"""

import os
import secrets

from web3 import Web3

from tests.integration.squid_issues.example_config import ExampleConfig
from ocean_keeper.account import Account
from squid_py import Ocean
from squid_py import ConfigProvider
from ocean_utils.ddo.ddo import DDO

def test_search_assets():

    # setup test tags to search
    search_tag = secrets.token_hex(32)

    config = ExampleConfig.get_config()
    ConfigProvider.set_config(config)
    ocn = Ocean()

    base_path = os.path.dirname(os.path.realpath(__file__))
    sample_ddo_path = os.path.join(base_path, 'ddo_sa_sample.json')
    assert os.path.exists(sample_ddo_path), "{} does not exist!".format(sample_ddo_path)

    accounts_config = ExampleConfig.get_accounts_config()
    ##########################################################
    # Setup account
    ##########################################################
    publisher = Account(
        Web3.toChecksumAddress(accounts_config['parity.address'].lower()),
        accounts_config['parity.password'],
        accounts_config['parity.keyfile']
    )

    ##########################################################
    # Create an asset DDO with valid metadata
    ##########################################################
    test_ddo = DDO(json_filename=sample_ddo_path)
    test_ddo.metadata['base']['tags']=['test', search_tag]
    test_ddo_dict = test_ddo.as_dictionary()
#    print(test_ddo_dict)
    asset = DDO(dictionary=test_ddo_dict)
    print(asset.metadata['base']['tags'])

    ##########################################################
    # Register using high-level interface
    ##########################################################
    ocn.assets.create(asset.metadata, publisher)

    query = {'query': { 'tags': [search_tag]} }
    ddo_list = ocn.assets.query(query)
    print('ddo list len', len(ddo_list))
    for ddo in ddo_list:
        print('found', ddo.metadata['base']['tags'])
        assert(search_tag in ddo.metadata['base']['tags'])
    assert(len(ddo_list) == 1)




"""
    Basic Purchase class
"""

from starfish.purchase.purchase_base import PurchaseBase
from starfish.exceptions import StarfishPurchaseError

class Purchase(PurchaseBase):
    """

    This class is returned by purchasing an asset uning the :func:`.Listing.purchase` method.

    :param agent: agent that was used create this object.
    :type agent: :class:`.Agent`
    :param listing: Listing used to purchase this asset.
    :type listing: :class:`.Listing`
    :param purchase_id: purchase_id used to buy this asset.
    :type purchase_id: str

    """

    def __init__(self, agent, listing, purchase_id, account):
        """init the the Purchase Object Base with the agent instance"""
        super().__init__(agent, listing, purchase_id, account)

    @property
    def is_completed(self):
        """

        Currently the same as `is_purchase_valid`, but renamed to be more meaningfull
            with the `wait_for_completion` method.

        :return: boolean True if the purchase has completed and finished, else
            False if the purchase is invalid or the has not finished.

        """
        return self.is_purchase_valid

    def wait_for_completion(self, timeout_seconds=60):
        """

        Some purchases ( squid ), require to wait for the smart contracts to complete
        This method will call the underlying agent to wait for the purchase to complete

        :param integer timeout_seconds: Optional seconds to waif to purchase to complete. Default: 60 seconds
        :return: True if successfull or an error message if failed
        :type: string or boolean

        :raises OceanPurchaseError: if the correct events are not received

        """
        if not self.is_purchased:
            raise StarfishPurchaseError('You need to purchase this asset before waiting')

        return self._agent.purchase_wait_for_completion(self._listing.asset, self._account, self._purchase_id, timeout_seconds)


    @property
    def consume_asset(self):
        """

        Consume a purchased asset. This call will try to download the asset data.

        You can call the :func:`is_purchased` property before hand to check that you
        have already purchased this asset.

        :param str download_path: Path to download the asset files too.

        :return: data returned from the asset , or False
        :type: object or False

        """
        if not self.is_purchased:
            return False

        return self._agent.consume_asset(self._listing, self._account, self._purchase_id)

    @property
    def get_type(self):
        return "asset"

    @property
    def is_purchased(self):
        """
        :return: True if this asset is a purchased asset.
        :type: boolean
        """
        return not (self._purchase_id is None or self._account is None)

    @property
    def is_purchase_valid(self):
        """

        Test to see if this purchased asset can be accessed and is valid.

        :return: boolean value if this asset has been purchased
        :type: boolean
        """
        if not self.is_purchased:
            return False

        return self._agent.is_access_granted_for_asset(self._listing.asset, self._account, self._purchase_id)




"""
    Base agent object
"""

from abc import ABC, abstractmethod
from starfish import Ocean

class AgentBase(ABC):
    """

    Base agent class

    :param ocean: Ocean object that is used by the Agent class
    :type ocean: :class:`.Ocean`
    """
    def __init__(self, ocean):
        """init the the Ocean Object Base with the ocean instance"""

        if not isinstance(ocean, Ocean):
            raise ValueError('You must pass a valid Ocean object')
        self._ocean = ocean

        super().__init__()

    @abstractmethod
    def register_asset(self, asset, listing_data, account):
        """

        Register a squid asset with the ocean network.

        :param asset: the Asset to register, at the moment only a Asset can be used.
        :type asset: :class:`.AssetBase` object to register
        :param dict listing_data: data providing listing information for the asset.
        :param account: Ocean account to use to register this asset.
        :type account: :class:`.Account` object to use for registration.

        :return: A new :class:`.Listing` object that has been registered, if failure then return None.
        :type: :class:`.Listing` class

        For example::

           # get your publisher account
           account = ocean.get_account('0x00bd138abd70e2f00903268f3db08f2d25677c9e')
           agent = SquidAgent(ocean)
           asset = FileAsset(filename='my_file.csv')
           listing_data = {
                'name': 'My data',
                'price': '4',
           }

           listing = agent.register_asset(asset, listing_data, account)

           if listing:
               print(f'registered my listing asset for sale with the did {listing.did}')

        """
        pass

    @abstractmethod
    def validate_asset(self, asset):
        """

        Validate an asset

        :param asset: Asset to validate.
        :return: True if the asset is valid
        """
        pass

    @abstractmethod
    def get_listing(self, listing_id):
        """

        Return an listing from the given listing_id.

        :param str listing_id: Id of the listing.

        :return: a registered listing given a Id of the listing
        :type: :class:`.Listing` class

        """
        pass

    @abstractmethod
    def search_listings(self, text, sort=None, offset=100, page=0):
        """

        Search the off chain storage for an asset with the givien 'text'

        :param str text: Test to search all metadata items for.
        :param sort: sort the results ( defaults: None, no sort).
        :type sort: str or None
        :param int offset: Return the result from with the maximum record count ( defaults: 100 ).
        :param int page: Returns the page number based on the offset.

        :return: a list of assets objects found using the search.
        :type: list of DID strings

        For example::

            # return the 300 -> 399 records in the search for the text 'weather' in the metadata.
            my_result = agent.search_registered_assets('weather', None, 100, 3)

        """
        pass

    @abstractmethod
    def purchase_asset(self, listing, account):
        """

        Purchase an asset using it's listing and an account.

        :param listing: Listing to use for the purchase.
        :type listing: :class:`.Listing`
        :param account: Ocean account to purchase the asset.
        :type account: :class:`.Account` object to use for registration.

        """
        pass

    @abstractmethod
    def is_access_granted_for_asset(self, asset, account, purchase_id=None):
        """

        Check to see if the account and purchase_id have access to the assed data.


        :param asset: Asset to check for access.
        :type asset: :class:`.Asset` object
        :param account: Ocean account to purchase the asset.
        :type account: :class:`.Account` object to use for registration.
        :param str purchase_id: purchase id that was used to purchase the asset.

        :return: True if the asset can be accessed and consumed.
        :type: boolean
        """
        pass

    @abstractmethod
    def get_asset_purchase_ids(self, asset):
        """

        Returns as list of purchase id's that have been used for this asset

        :param asset: Asset to return purchase details.
        :type asset: :class:`.Asset` object

        :return: list of purchase ids
        :type: list

        """
        pass

    @abstractmethod
    def purchase_wait_for_completion(self, asset, account,  purchase_id, timeoutSeconds):
        """

            Wait for completion of the purchase

            TODO: issues here...
            + No method as yet to pass back paramaters and values during the purchase process
            + We assume that the following templates below will always be used.

        """
        pass

    @abstractmethod
    def consume_asset(self, listing, account, purchase_id):
        """
        Consume the asset and download the data. The actual payment to the asset
        provider will be made at this point.

        :param listing: Listing that was used to make the purchase.
        :type listing: :class:`.Listing`
        :param account: Ocean account that was used to purchase the asset.
        :type account: :class:`.Account` object to use for registration.
        :param str purchase_id: purchase id that was used to purchase the asset.

        :return: True if the asset has been consumed and downloaded
        :type: boolean

        """

    @property
    def ocean(self):
        """
        :return: Ocean object
        :type: :class:`.Ocean`
        """
        return self._ocean



"""
    test_08_asset_registration

    As a developer working with Ocean,
    I need a way to register a new asset with Ocean

"""

import secrets
import json

from starfish.asset import DataAsset
from starfish.agent import RemoteAgent
from starfish.ddo.starfish_ddo import StarfishDDO



def test_08_asset_registration(resources, remote_agent):
    test_data = secrets.token_hex(1024)
    asset1 = DataAsset.create('TestAsset', test_data)
    listing1 = remote_agent.register_asset(asset1, resources.listing_data)
    assert(listing1)
    assert(listing1.asset)

    asset2 = DataAsset.create('TestAsset', test_data)
    listing2 = remote_agent.register_asset(asset2, resources.listing_data)
    assert(listing2)
    assert(listing2.asset)

    assert(listing1.asset.asset_id == listing2.asset.asset_id)



"""

Memory Agent class to provide basic functionality for Ocean Agents

"""

import secrets
import re
import json

from ocean_utils.did import id_to_did, did_to_id

from starfish.agent import AgentBase
from starfish.listing import Listing
from starfish.purchase import Purchase
from starfish.utils.did import did_parse


class MemoryAgent(AgentBase):
    """

    Memory Agent class allows to register, list, purchase and consume assets.

    :param ocean: Ocean object that is being used.
    :type ocean: :class:`.Ocean`

    """

    def __init__(self, ocean, *args, **kwargs):
        """init a standard ocean object"""
        AgentBase.__init__(self, ocean)

        if args and isinstance(args[0], dict):
            kwargs = args[0]

        self._memory = {
            'listing_data': {},
            'asset': {},
            'purchase': {}
        }

    def register_asset(self, asset, listing_data, account=None):
        """

        Register a memory asset with the ocean network.

        :param dict metadata: metadata dictionary to store for this asset.
        :param account: Optional, since an account is not assigned to an registered memory asset.
        :type account: :class:`.Account` object to use for registration.

        :return: A new :class:`.Listing` object that has been registered, if failure then return None.
        :type: :class:`.Listing` class

        """

        asset_did = id_to_did(secrets.token_hex(64))
        listing_id = did_to_id(asset_did)
        listing_data['listing_id'] = listing_id,
        listing_data['asset_did'] = asset_did

        self._memory['listing_data'][listing_id] = listing_data
        self._memory['asset'][asset_did] = asset

        listing = None
        if listing_data:
            asset.set_did(asset_did)
            listing = Listing(self, listing_id, asset, listing_data)

        return listing

    def validate_asset(self, asset):
        """

        Validate an asset

        :param asset: Asset to validate.
        :return: True if the asset is valid
        """
        return asset is not None


    def get_listing(self, listing_id):
        """

        Return an listing from the given listing_id.

        :param str listing_id: Id of the listing.

        :return: a registered listing given a Id of the listing
        :type: :class:`.Listing` class

        """
        listing = None
        if listing_id in self._memory['listing_data']:
            listing_data = self._memory['listing_data'][listing_id]
            if listing_data:
                asset_did = listing_data['asset_did']
                asset = self._memory['asset'][asset_did]
                listing = Listing(self, listing_id, asset, listing_data)

        return listing


    def search_listings(self, text, sort=None, offset=100, page=0):
        """

        Search for listings with the givien 'text'

        :param str text: Text to search all listing data for.
        :param sort: sort the results ( defaults: None, no sort).
        :type sort: str or None
        :param int offset: Return the result from with the maximum record count ( defaults: 100 ).
        :param int page: Returns the page number based on the offset.

        :return: a list of listing objects found using the search.
        :type: :class:`.Listing` objects

        For example::

            # return the 300 -> 399 records in the search for the text 'weather' in the metadata.
            my_result = agent.search_registered_assets('weather', None, 100, 3)

        """
        listing_items = None
        for listing_id, listing_data in self._memory['listing_data'].items():
            if re.search(text, json.dumps(listing_data)):
                if listing_items is None:
                    listing_items = {}
                listing_items[listing_id] = listing_data

        return listing_items

    def purchase_asset(self, listing, account):
        """

        Purchase an asset using it's listing and an account.

        :param listing: Listing to use for the purchase.
        :type listing: :class:`.Listing`
        :param account: Ocean account to purchase the asset.
        :type account: :class:`.Account` object to use for registration.

        """
        purchase = None

        purchase_id = secrets.token_hex(64)
        if purchase_id:
            purchase = Purchase(self, listing, purchase_id, account)
            self._memory['purchase'][purchase_id] = (purchase, account.address)

        return purchase

    def purchase_wait_for_completion(self, asset, account,  purchase_id, timeoutSeconds):
        """

            Wait for completion of the purchase

            TODO: issues here...
            + No method as yet to pass back paramaters and values during the purchase process
            + We assume that the following templates below will always be used.

        """
        return True
    def is_access_granted_for_asset(self, asset, account, purchase_id=None):
        """

        Check to see if the account and purchase_id have access to the assed data.


        :param asset: Asset to check for access.
        :type asset: :class:`.Asset` object
        :param account: Ocean account to purchase the asset.
        :type account: :class:`.Account` object to use for registration.
        :param str purchase_id: purchase id that was used to purchase the asset.

        :return: True if the asset can be accessed and consumed.
        :type: boolean
        """

        if purchase_id in self._memory['purchase']:
            purchase, account_address = self._memory['purchase'][purchase_id]
            return purchase and account.is_address_equal(account_address)

        return False

    def get_asset_purchase_ids(self, asset):
        """

        Returns as list of purchase id's that have been used for this asset

        :param asset: DataAsset to return purchase details.
        :type asset: :class:`.DataAsset` object

        :return: list of purchase ids
        :type: list

        """
        return []

    def consume_asset(self, listing, account, purchase_id):
        """
        Consume the asset and download the data. The actual payment to the asset
        provider will be made at this point.

        :param listing: Listing that was used to make the purchase.
        :type listing: :class:`.Listing`
        :param str purchase_id: purchase id that was used to purchase the asset.
        :param account: Ocean account that was used to purchase the asset.
        :type account: :class:`.Account` object to use for registration.

        :return: True if the asset has been consumed and downloaded
        :type: boolean

        """
        return purchase_id in self._memory['purchase']

    @staticmethod
    def is_did_valid(did):
        """
        Checks to see if the DID string is a valid DID for this type of Asset.
        This method only checks the syntax of the DID, it does not resolve the DID
        to see if it is assigned to a valid Asset.

        :param str did: DID string to check to see if it is in a valid format.

        :return: True if the DID is in the format 'did:dep:xxxxx'
        :type: boolean
        """
        data = did_parse(did)
        return not data['path']



"""
    Metadata class to handle asset metadata.


"""

# from starfish import logger

class Metadata():
    """

    :param agent: agent object to used to create
    :type agent: :class:`.AgentObject`
    :param metadata: Metadata.
    :type metadata: dict

    """
    def __init__(self, name=None, asset_type=None, **kwargs):
        """
        init the metadata class
        """

        self._valid_fields = [
            'name', 'type', 'description', 'dateCreated',
            'author', 'license', 'copyrightHolder', 'links', 'inLanguage',
            'tags', 'additionalInformation', 'files'
        ]

        self._data = {}

        if isinstance(name, dict):
            for name_value, value in name.items():
                self.__setitem__(name_value, value)
        elif isinstance(name, str):
            self.name = name
        else:
            raise ValueError('You need to pass the meta name as a string or dict')

        self.asset_type = asset_type

    def __setitem__(self, name, value):
        if name in self._valid_fields:
            self._data[name] = value

    def __getitem__(self, name):
        if name in self._data:
            return self._data[name]
        return None



"""
    Test ocean class

"""

import datetime
import pathlib
import json
import logging
import time
import secrets
from web3 import Web3

from starfish import Ocean
from starfish.agent import MemoryAgent
from starfish.asset import DataAsset


def _register_asset_for_sale(agent, resources, account):

    asset = DataAsset.create('TestAsset', secrets.token_hex(256))
    listing = agent.register_asset(asset, resources.listing_data, account=account)
    assert listing
    assert listing.asset.did
    return listing

def test_asset(ocean, resources, config, publisher_account, purchaser_account):

    # create an ocean object

    agent = MemoryAgent(ocean)
    assert agent


    listing = _register_asset_for_sale(agent, resources, publisher_account)
    assert listing
    assert publisher_account

    listing_id = listing.listing_id
    # start to test getting the asset from storage
    listing = agent.get_listing(listing_id)
    assert listing
    assert listing.listing_id == listing_id


    # test purchase an asset
    purchase_asset = listing.purchase(purchaser_account)
    assert purchase_asset


    assert(purchase_asset.is_purchased)
    assert(purchase_asset.wait_for_completion())
    assert(purchase_asset.is_completed)
    assert(purchase_asset.is_purchase_valid)

    assert(purchase_asset.consume_asset)



def test_search_listing(ocean, resources, config, publisher_account):

    agent = MemoryAgent(ocean)

    listing = _register_asset_for_sale(agent, resources, publisher_account)
    assert listing
    assert publisher_account

    # choose a word from the description field
    text = resources.listing_data['author']
    words = text.split(' ')
    word = words[0]

    # should return at least 1 or more assets
    logging.info(f'search word is {word}')
    searchResult = agent.search_listings(word)
    assert searchResult

    assert(len(searchResult) > 0)




import logging
import re
import os
import sys

logger = logging.getLogger('starfish.artifacts')

def is_contract_type_exists(network_name, path):
    for filename in os.listdir(path):
        if re.match(fr'\w+\.{network_name}\.json', filename):
            return True
    return False

def find_contract_path(network_name):
    folder = os.path.abspath('artifacts')
    # first search the working folder of the app
    logger.debug(f'search for artifacts {folder}')
    if os.path.exists(folder):
        if is_contract_type_exists(network_name, folder):
            return folder
    logger.debug(f'search for artifacts {folder}')
    folder = os.path.join(sys.prefix, 'artifacts')
    if os.path.exists(folder):
        if is_contract_type_exists(network_name, folder):
            return folder

    logger.debug(f'artifacts not found')
    return None
